#!/usr/bin/env python
"""A Storage SCU application.

Used for transferring DICOM SOP Instances to a Storage SCP.
"""

import argparse
import os
from pathlib import Path
import sys

from pydicom import dcmread
from pydicom.errors import InvalidDicomError
from pydicom.uid import (
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    ExplicitVRBigEndian,
    DeflatedExplicitVRLittleEndian,
)

from pynetdicom import AE, StoragePresentationContexts
from pynetdicom.apps.common import setup_logging, get_files
from pynetdicom._globals import DEFAULT_MAX_LENGTH


__version__ = "0.3.0"


def _setup_argparser():
    """Setup the command line arguments"""
    # Description
    parser = argparse.ArgumentParser(
        description=(
            "The storescu application implements a Service Class User "
            "(SCU) for the Storage Service Class. For each DICOM "
            "file on the command line it sends a C-STORE message to a "
            "Storage Service Class Provider (SCP) and waits for a response."
        ),
        usage="storescu [options] addr port path",
    )

    # Parameters
    req_opts = parser.add_argument_group("Parameters")
    req_opts.add_argument(
        "addr", help="TCP/IP address or hostname of DICOM peer", type=str
    )
    req_opts.add_argument("port", help="TCP/IP port number of peer", type=int)
    req_opts.add_argument(
        "path",
        metavar="path",
        nargs="+",
        help="DICOM file or directory to be transmitted",
        type=str,
    )

    # General Options
    gen_opts = parser.add_argument_group("General Options")
    gen_opts.add_argument(
        "--version", help="print version information and exit", action="store_true"
    )
    output = gen_opts.add_mutually_exclusive_group()
    output.add_argument(
        "-q",
        "--quiet",
        help="quiet mode, print no warnings and errors",
        action="store_const",
        dest="log_type",
        const="q",
    )
    output.add_argument(
        "-v",
        "--verbose",
        help="verbose mode, print processing details",
        action="store_const",
        dest="log_type",
        const="v",
    )
    output.add_argument(
        "-d",
        "--debug",
        help="debug mode, print debug information",
        action="store_const",
        dest="log_type",
        const="d",
    )
    gen_opts.add_argument(
        "-ll",
        "--log-level",
        metavar="[l]",
        help=("use level l for the logger (critical, error, warn, info, debug)"),
        type=str,
        choices=["critical", "error", "warn", "info", "debug"],
    )

    # Input Options
    in_opts = parser.add_argument_group("Input Options")
    in_opts.add_argument(
        "-r",
        "--recurse",
        help="recursively search the given directory",
        action="store_true",
    )

    # Network Options
    net_opts = parser.add_argument_group("Network Options")
    net_opts.add_argument(
        "-aet",
        "--calling-aet",
        metavar="[a]etitle",
        help="set my calling AE title (default: STORESCU)",
        type=str,
        default="STORESCU",
    )
    net_opts.add_argument(
        "-aec",
        "--called-aet",
        metavar="[a]etitle",
        help="set called AE title of peer (default: ANY-SCP)",
        type=str,
        default="ANY-SCP",
    )
    net_opts.add_argument(
        "-ta",
        "--acse-timeout",
        metavar="[s]econds",
        help="timeout for ACSE messages (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-td",
        "--dimse-timeout",
        metavar="[s]econds",
        help="timeout for DIMSE messages (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-tn",
        "--network-timeout",
        metavar="[s]econds",
        help="timeout for the network (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-pdu",
        "--max-pdu",
        metavar="[n]umber of bytes",
        help=(
            f"set max receive pdu to n bytes (0 for unlimited, "
            f"default: {DEFAULT_MAX_LENGTH})"
        ),
        type=int,
        default=DEFAULT_MAX_LENGTH,
    )

    # Transfer Syntaxes
    ts_opts = parser.add_argument_group("Transfer Syntax Options")
    syntax = ts_opts.add_mutually_exclusive_group()
    syntax.add_argument(
        "-xe",
        "--request-little",
        help="request explicit VR little endian TS only",
        action="store_true",
    )
    syntax.add_argument(
        "-xb",
        "--request-big",
        help="request explicit VR big endian TS only",
        action="store_true",
    )
    syntax.add_argument(
        "-xi",
        "--request-implicit",
        help="request implicit VR little endian TS only",
        action="store_true",
    )

    # Misc Options
    misc_opts = parser.add_argument_group("Miscellaneous Options")
    misc_opts.add_argument(
        "-cx",
        "--required-contexts",
        help=(
            "only request the presentation contexts required for the "
            "input DICOM file(s)"
        ),
        action="store_true",
    )

    return parser.parse_args()


def get_contexts(fpaths, app_logger):
    """Return the valid DICOM files and their context values.

    Parameters
    ----------
    fpaths : list of str
        A list of paths to the files to try and get data from.

    Returns
    -------
    list of str, dict
        A list of paths to valid DICOM files and the {SOP Class UID :
        [Transfer Syntax UIDs]} that can be used to create the required
        presentation contexts.
    """
    good, bad = [], []
    contexts = {}
    for fpath in fpaths:
        path = os.fspath(Path(fpath).resolve())
        try:
            ds = dcmread(path)
        except Exception:
            bad.append(("Bad DICOM file", path))
            continue

        try:
            sop_class = ds.SOPClassUID
            tsyntax = ds.file_meta.TransferSyntaxUID
        except Exception:
            bad.append(("Unknown SOP Class or Transfer Syntax UID", path))
            continue

        tsyntaxes = contexts.setdefault(sop_class, [])
        if tsyntax not in tsyntaxes:
            tsyntaxes.append(tsyntax)

        good.append(path)

    for reason, path in bad:
        app_logger.error(f"{reason}: {path}")

    return good, contexts


def main(args=None):
    """Run the application."""
    if args is not None:
        sys.argv = args

    args = _setup_argparser()

    if args.version:
        print(f"storescu.py v{__version__}")
        sys.exit()

    APP_LOGGER = setup_logging(args, "storescu")
    APP_LOGGER.debug(f"storescu.py v{__version__}")
    APP_LOGGER.debug("")

    lfiles, badfiles = get_files(args.path, args.recurse)

    for bad in badfiles:
        APP_LOGGER.error(f"Cannot access path: {bad}")

    ae = AE(ae_title=args.calling_aet)
    ae.acse_timeout = args.acse_timeout
    ae.dimse_timeout = args.dimse_timeout
    ae.network_timeout = args.network_timeout

    if args.required_contexts:
        # Only propose required presentation contexts
        lfiles, contexts = get_contexts(lfiles, APP_LOGGER)
        try:
            for abstract, transfer in contexts.items():
                for tsyntax in transfer:
                    ae.add_requested_context(abstract, tsyntax)
        except ValueError:
            raise ValueError(
                "More than 128 presentation contexts required with "
                "the '--required-contexts' flag, please try again "
                "without it or with fewer files"
            )
    else:
        # Propose the default presentation contexts
        if args.request_little:
            transfer_syntax = [ExplicitVRLittleEndian]
        elif args.request_big:
            transfer_syntax = [ExplicitVRBigEndian]
        elif args.request_implicit:
            transfer_syntax = [ImplicitVRLittleEndian]
        else:
            transfer_syntax = [
                ExplicitVRLittleEndian,
                ImplicitVRLittleEndian,
                DeflatedExplicitVRLittleEndian,
                ExplicitVRBigEndian,
            ]

        for cx in StoragePresentationContexts:
            ae.add_requested_context(cx.abstract_syntax, transfer_syntax)

    if not lfiles:
        APP_LOGGER.warning("No suitable DICOM files found")
        sys.exit()

    # Request association with remote
    assoc = ae.associate(
        args.addr, args.port, ae_title=args.called_aet, max_pdu=args.max_pdu
    )
    if assoc.is_established:
        ii = 1
        for fpath in lfiles:
            APP_LOGGER.info(f"Sending file: {fpath}")
            try:
                ds = dcmread(fpath)
                assoc.send_c_store(ds, ii)
                ii += 1
            except InvalidDicomError:
                APP_LOGGER.error(f"Bad DICOM file: {fpath}")
            except Exception as exc:
                APP_LOGGER.error(f"Store failed: {fpath}")
                APP_LOGGER.exception(exc)

        assoc.release()
    else:
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
The main user class, represents a DICOM Application Entity
"""

from copy import deepcopy
from datetime import datetime
import logging
from ssl import SSLContext
import threading
from typing import (
    Union,
    Optional,
    List,
    Tuple,
    Dict,
    cast,
    TypeVar,
    Type,
    Any,
    Sequence,
)
import warnings

from pydicom.uid import UID

from pynetdicom import _config
from pynetdicom.association import Association
from pynetdicom.events import EventHandlerType
from pynetdicom.presentation import PresentationContext
from pynetdicom.pdu_primitives import _UI
from pynetdicom.transport import (
    AssociationSocket,
    AssociationServer,
    ThreadedAssociationServer,
)
from pynetdicom.utils import make_target, set_ae, decode_bytes, set_uid
from pynetdicom._globals import (
    MODE_REQUESTOR,
    DEFAULT_MAX_LENGTH,
    DEFAULT_TRANSFER_SYNTAXES,
    BIND_ADDRESS,
)


LOGGER = logging.getLogger(__name__)


_T = TypeVar("_T")
ListCXType = list[PresentationContext]
TSyntaxType = None | str | UID | Sequence[str] | Sequence[UID]


class ApplicationEntity:
    """Represents a DICOM Application Entity (AE).

    An AE may be a *Service Class Provider* (SCP), a *Service Class User* (SCU)
    or both.
    """

    # pylint: disable=too-many-instance-attributes,too-many-public-methods
    def __init__(self, ae_title: str = "PYNETDICOM") -> None:
        """Create a new Application Entity.

        .. versionchanged:: 2.0

            `ae_title` should be :class:`str`

        Parameters
        ----------
        ae_title : str, optional
            The AE title of the Application Entity as an ASCII string
            (default: ``'PYNETDICOM'``).
        """
        self._ae_title: str
        self.ae_title = ae_title

        from pynetdicom import (
            PYNETDICOM_IMPLEMENTATION_UID,
            PYNETDICOM_IMPLEMENTATION_VERSION,
        )

        # Default Implementation Class UID and Version Name
        self._implementation_uid: UID = PYNETDICOM_IMPLEMENTATION_UID
        self._implementation_version: Optional[str] = PYNETDICOM_IMPLEMENTATION_VERSION

        # List of PresentationContext
        self._requested_contexts: ListCXType = []
        # {abstract_syntax : PresentationContext}
        self._supported_contexts: Dict[UID, PresentationContext] = {}

        # Default maximum simultaneous associations
        self._maximum_associations = 10

        # Default maximum PDU receive size (in bytes)
        self._maximum_pdu_size = DEFAULT_MAX_LENGTH

        # Default timeouts - None means no timeout
        self._acse_timeout: Optional[float] = 30
        self._connection_timeout: Optional[float] = None
        self._dimse_timeout: Optional[float] = 30
        self._network_timeout: Optional[float] = 60

        # Require Calling/Called AE titles to match if value is non-empty str
        self._require_calling_aet: List[str] = []
        self._require_called_aet = False

        self._servers: List[ThreadedAssociationServer] = []
        self._lock: threading.Lock = threading.Lock()

    @property
    def acse_timeout(self) -> Optional[float]:
        """Get or set the ACSE timeout value (in seconds).

        Parameters
        ----------
        value : Union[int, float, None]
            The maximum amount of time (in seconds) to wait for association
            related messages. A value of ``None`` means no timeout. (default:
            ``30``)
        """
        return self._acse_timeout

    @acse_timeout.setter
    def acse_timeout(self, value: Optional[float]) -> None:
        """Set the ACSE timeout (in seconds)."""
        if value is None:
            self._acse_timeout = None
        elif isinstance(value, (int, float)) and value >= 0:
            self._acse_timeout = value
        else:
            LOGGER.warning("ACSE timeout set to 30 seconds")
            self._acse_timeout = 30

        for assoc in self.active_associations:
            assoc.acse_timeout = self.acse_timeout

    @property
    def active_associations(self) -> List[Association]:
        """Return a list of the AE's active
        :class:`~pynetdicom.association.Association` threads.

        Returns
        -------
        list of Association
            A list of all active association threads, both requestors and
            acceptors.
        """
        threads = threading.enumerate()
        t_assocs = [tt for tt in threads if isinstance(tt, Association)]

        return [tt for tt in t_assocs if tt.ae == self]

    def add_requested_context(
        self,
        abstract_syntax: Union[str, UID],
        transfer_syntax: TSyntaxType = None,
    ) -> None:
        """Add a :ref:`presentation context<user_presentation>` to be
        proposed when requesting an association.

        When an SCU sends an association request to a peer it includes a list
        of presentation contexts it would like the peer to support. This
        method adds a single
        :class:`~pynetdicom.presentation.PresentationContext` to the list of
        the SCU's requested contexts.

        Only 128 presentation contexts can be included in the association
        request. Multiple presentation contexts may be requested with the
        same abstract syntax.

        To remove a requested context or one or more of its transfer syntaxes
        see the :meth:`remove_requested_context` method.

        Parameters
        ----------
        abstract_syntax : str or pydicom.uid.UID
            The abstract syntax of the presentation context to request.
        transfer_syntax :  str/pydicom.uid.UID or list of str/pydicom.uid.UID
            The transfer syntax(es) to request (default:
            :attr:`~pynetdicom._globals.DEFAULT_TRANSFER_SYNTAXES`).

        Raises
        ------
        ValueError
            If 128 requested presentation contexts have already been added.

        Examples
        --------
        Add a requested presentation context for *Verification SOP Class* with
        the default transfer syntaxes by using its UID value.

        >>> from pynetdicom import AE
        >>> ae = AE()
        >>> ae.add_requested_context('1.2.840.10008.1.1')
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Little Endian
            =Explicit VR Big Endian

        Add a requested presentation context for *Verification SOP Class* with
        the default transfer syntaxes by using the inbuilt
        :class:`~pynetdicom.sop_class.Verification` object.

        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_requested_context(Verification)

        Add a requested presentation context for *Verification SOP Class* with
        a transfer syntax of *Implicit VR Little Endian*.

        >>> from pydicom.uid import ImplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_requested_context(Verification, ImplicitVRLittleEndian)
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian

        Add two requested presentation contexts for *Verification SOP Class*
        using different transfer syntaxes for each.

        >>> from pydicom.uid import (
        ...     ImplicitVRLittleEndian, ExplicitVRLittleEndian, ExplicitVRBigEndian
        ... )
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_requested_context(
        ...     Verification, [ImplicitVRLittleEndian, ExplicitVRBigEndian]
        ... )
        >>> ae.add_requested_context(Verification, ExplicitVRLittleEndian)
        >>> len(ae.requested_contexts)
        2
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Big Endian
        >>> print(ae.requested_contexts[1])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Explicit VR Little Endian

        References
        ----------

        * DICOM Standard, Part 8,
          :dcm:`Section 7.1.1.13<part08.html#sect_7.1.1.13>`
        * DICOM Standard, Part 8,
          :dcm:`Table 9-18<part08.html#table_9-18>`
        """
        if transfer_syntax is None:
            transfer_syntax = DEFAULT_TRANSFER_SYNTAXES

        if len(self.requested_contexts) >= 128:
            raise ValueError(
                "Failed to add the requested presentation context as there "
                "are already the maximum allowed number of requested contexts"
            )

        # Allow single transfer syntax values for convenience
        if isinstance(transfer_syntax, str):
            transfer_syntax = [transfer_syntax]

        context = PresentationContext()
        context.abstract_syntax = UID(abstract_syntax)
        context.transfer_syntax = [UID(syntax) for syntax in transfer_syntax]

        self._requested_contexts.append(context)

    def add_supported_context(
        self,
        abstract_syntax: Union[str, UID],
        transfer_syntax: TSyntaxType = None,
        scu_role: Optional[bool] = None,
        scp_role: Optional[bool] = None,
    ) -> None:
        """Add a :ref:`presentation context<user_presentation>` to be
        supported when accepting association requests.

        When an association request is received from a peer it supplies a list
        of presentation contexts that it would like the SCP to support. This
        method adds a :class:`~pynetdicom.presentation.PresentationContext`
        to the list of the SCP's supported contexts.

        Where the abstract syntax is already supported the transfer syntaxes
        will be extended by those supplied in `transfer_syntax`. To remove
        a supported context or one or more of its transfer syntaxes see the
        :meth:`remove_supported_context` method.

        Parameters
        ----------
        abstract_syntax : str, pydicom.uid.UID
            The abstract syntax of the presentation context to be supported.
        transfer_syntax :  str/pydicom.uid.UID or list of str/pydicom.uid.UID
            The transfer syntax(es) to support (default:
            :attr:`~pynetdicom._globals.DEFAULT_TRANSFER_SYNTAXES`).
        scu_role : bool or None, optional
            If the association requestor includes an
            :ref:`SCP/SCU Role Selection Negotiation<user_presentation_role>`
            item for this context then:

            * If ``None`` then ignore the proposal (if either `scp_role` or
              `scu_role` is ``None`` then both are assumed to be) and use the
              default roles.
            * If ``True`` accept the proposed SCU role
            * If ``False`` reject the proposed SCU role
        scp_role : bool or None, optional
            If the association requestor includes an
            :ref:`SCP/SCU Role Selection Negotiation<user_presentation_role>`
            item for this context then:

            * If ``None`` then ignore the proposal (if either `scp_role` or
              `scu_role` is ``None`` then both are assumed to be) and use the
              default roles.
            * If ``True`` accept the proposed SCP role
            * If ``False`` reject the proposed SCP role

        Examples
        --------
        Add support for presentation contexts with an abstract syntax of
        *Verification SOP Class* and the default transfer syntaxes by using
        its UID value.

        >>> from pynetdicom import AE
        >>> ae = AE()
        >>> ae.add_supported_context('1.2.840.10008.1.1')
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Little Endian
            =Explicit VR Big Endian

        Add support for presentation contexts with an abstract syntax of
        *Verification SOP Class* and the default transfer syntaxes by using the
        inbuilt :class:`~pynetdicom.sop_class.Verification` object.

        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_supported_context(Verification)

        Add support for presentation contexts with an abstract syntax of
        *Verification SOP Class* and a transfer syntax of *Implicit VR Little
        Endian*.

        >>> from pydicom.uid import ImplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_supported_context(Verification, ImplicitVRLittleEndian)
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian

        Add support for presentation contexts with an abstract syntax of
        *Verification SOP Class* and transfer syntaxes of *Implicit VR Little
        Endian* and *Explicit VR Big Endian* and then update the context to
        also support *Explicit VR Little Endian*.

        >>> from pydicom.uid import (
        ...     ImplicitVRLittleEndian, ExplicitVRLittleEndian, ExplicitVRBigEndian
        ... )
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_supported_context(
        ...     Verification, [ImplicitVRLittleEndian, ExplicitVRBigEndian]
        ... )
        >>> ae.add_supported_context(Verification, ExplicitVRLittleEndian)
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Big Endian
            =Explicit VR Little Endian

        Add support for *CT Image Storage* and if the association requestor
        includes an SCP/SCU Role Selection Negotiation item for *CT Image
        Storage* requesting the SCU and SCP roles then accept the proposal.

        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import CTImageStorage
        >>> ae = AE()
        >>> ae.add_supported_context(CTImageStorage, scu_role=True, scp_role=True)
        """
        if transfer_syntax is None:
            transfer_syntax = DEFAULT_TRANSFER_SYNTAXES  # List[str]

        abstract_syntax = UID(abstract_syntax)

        if not isinstance(scu_role, (type(None), bool)):
            raise TypeError("`scu_role` must be None or bool")

        if not isinstance(scp_role, (type(None), bool)):
            raise TypeError("`scp_role` must be None or bool")

        # For convenience allow single transfer syntax values
        if isinstance(transfer_syntax, str):
            transfer_syntax = [transfer_syntax]

        transfer_syntax = [UID(ts) for ts in transfer_syntax]

        # If the abstract syntax is already supported then update the transfer
        #   syntaxes
        if abstract_syntax in self._supported_contexts:
            context = self._supported_contexts[abstract_syntax]
            for syntax in transfer_syntax:
                context.add_transfer_syntax(syntax)
            context.scu_role = None or scu_role
            context.scp_role = None or scp_role
        else:
            context = PresentationContext()
            context.abstract_syntax = abstract_syntax
            context.transfer_syntax = transfer_syntax  # type: ignore
            context.scu_role = None or scu_role
            context.scp_role = None or scp_role

            self._supported_contexts[abstract_syntax] = context

    @property
    def ae_title(self) -> str:
        """Get or set the AE title as :class:`str`.

        .. versionchanged:: 2.0

            `ae_title` should be set using :class:`str` and returns
            :class:`str` rather than :class:`bytes`

        Parameters
        ----------
        value : str
            The AE title to use for the local Application Entity as an ASCII
            string.

        Returns
        -------
        str
            The local Application Entity's AE title.
        """
        return self._ae_title

    @ae_title.setter
    def ae_title(self, value: str) -> None:
        """Set the AE title using :class:`str`."""
        if isinstance(value, bytes):
            warnings.warn(
                "The use of bytes with 'ae_title' is deprecated, use an ASCII "
                "str instead",
                DeprecationWarning,
            )
            value = decode_bytes(value)

        self._ae_title = cast(str, set_ae(value, "ae_title", False, False))

    def associate(
        self,
        addr: str,
        port: int,
        contexts: Optional[ListCXType] = None,
        ae_title: str = "ANY-SCP",
        max_pdu: int = DEFAULT_MAX_LENGTH,
        ext_neg: Optional[List[_UI]] = None,
        bind_address: Tuple[str, int] = BIND_ADDRESS,
        tls_args: Optional[Tuple[SSLContext, str]] = None,
        evt_handlers: Optional[List[EventHandlerType]] = None,
    ) -> Association:
        """Request an association with a remote AE.

        An :class:`~pynetdicom.association.Association` thread is returned
        whether or not the association is accepted and should be checked using
        :attr:`Association.is_established
        <pynetdicom.association.Association.is_established>`
        before sending any messages. The returned thread will only be running
        if the association was established.

        .. versionchanged:: 1.2

            Added `bind_address` and `tls_arg` keyword parameters

        .. versionchanged:: 1.3

            Added `evt_handlers` keyword parameter

        .. versionchanged:: 1.5

            `evt_handlers` now takes a list of 2- or 3-tuples

        .. versionchanged:: 2.0

            * `ae_title` should now be :class:`str`

        Parameters
        ----------
        addr : str
            The peer AE's TCP/IP address.
        port : int
            The peer AE's listen port number.
        contexts : list of presentation.PresentationContext, optional
            The presentation contexts that will be requested by the AE for
            support by the peer. If not used then the presentation contexts in
            the :attr:`requested_contexts` property will be requested instead.
        ae_title : str, optional
            The peer's AE title, will be used as the *Called AE Title*
            parameter value (default ``'ANY-SCP'``).
        max_pdu : int, optional
            The :dcm:`maximum PDV receive size<part08/chapter_D.html#sect_D.1>`
            in bytes to use when negotiating the association (default
            ``16832``). A value of ``0`` means the PDV size is unlimited.
        ext_neg : list of UserInformation objects, optional
            A list containing optional extended negotiation items:

            .. currentmodule:: pynetdicom.pdu_primitives

            * :class:`AsynchronousOperationsWindowNegotiation` (0 or 1 item)
            * :class:`~SCP_SCU_RoleSelectionNegotiation` (0 to N items)
            * :class:`~SOPClassCommonExtendedNegotiation` (0 to N items)
            * :class:`~SOPClassExtendedNegotiation` (0 to N items)
            * :class:`~UserIdentityNegotiation` (0 or 1 item)
        bind_address : 2-tuple, optional
            The (host, port) to bind the association's communication socket
            to, default ``("", 0)``.
        tls_args : 2-tuple, optional
            If TLS is required then this should be a 2-tuple containing a
            (`ssl_context`, `server_hostname`), where `ssl_context` is the
            :class:`ssl.SSLContext` instance to use to wrap the client socket
            and `server_hostname` is the value to use for the corresponding
            keyword argument in :meth:`~ssl.SSLContext.wrap_socket`. If no
            `tls_args` is supplied then TLS will not be used (default).
        evt_handlers : list of 2- or 3-tuple, optional
            A list of (*event*, *handler*) or (*event*, *handler*, *args*),
            where `event` is an ``evt.EVT_*`` event tuple, `handler` is a
            callable function that will be bound to the event and `args` is a
            :class:`list` of objects that will be passed to `handler` as
            optional extra arguments. At a minimum, `handler` should take an
            :class:`~pynetdicom.events.Event` parameter and may return or yield
            objects depending on the exact event that the handler is bound to.
            For more information see the :ref:`documentation<user_events>`.

        Returns
        -------
        assoc : association.Association
            If the association was established then a running
            :class:`~pynetdicom.association.Association` thread, otherwise
            returns a thread that hasn't been started.

        Raises
        ------
        RuntimeError
            If called with no requested presentation contexts (i.e. `contexts`
            has not been supplied and
            :attr:`~pynetdicom.ae.ApplicationEntity.requested_contexts` is
            empty).
        """
        if not isinstance(addr, str):
            raise TypeError("'addr' must be a valid IPv4 string")

        if not isinstance(port, int):
            raise TypeError("'port' must be a valid port number")

        # Association
        assoc = Association(self, MODE_REQUESTOR)

        # Set the thread name
        timestamp = datetime.strftime(datetime.now(), "%Y%m%d%H%M%S")
        assoc.name = f"RequestorThread@{timestamp}"

        # Setup the association's communication socket
        sock = self._create_socket(assoc, bind_address, tls_args)
        assoc.set_socket(sock)

        # Association Acceptor object -> remote AE
        # `ae_title` validation is performed by the ServiceUser
        assoc.acceptor.ae_title = ae_title
        assoc.acceptor.address = addr
        assoc.acceptor.port = port

        # Association Requestor object -> local AE
        assoc.requestor.address = sock.get_local_addr()
        assoc.requestor.port = bind_address[1]
        assoc.requestor.ae_title = self.ae_title
        assoc.requestor.maximum_length = max_pdu
        assoc.requestor.implementation_class_uid = self.implementation_class_uid
        assoc.requestor.implementation_version_name = self.implementation_version_name
        for item in ext_neg or []:
            assoc.requestor.add_negotiation_item(item)

        # Requestor's presentation contexts
        contexts = contexts or self.requested_contexts
        self._validate_requested_contexts(contexts)

        # PS3.8 Table 9.11, an A-ASSOCIATE-RQ must contain one or more
        #   Presentation Context items
        if not contexts:
            raise RuntimeError(
                "At least one requested presentation context is required "
                "before associating with a peer"
            )

        # Set using a copy of the original to play nicely
        contexts = deepcopy(contexts)

        # Add the context IDs
        for ii, context in enumerate(contexts):
            context.context_id = 2 * ii + 1

        assoc.requestor.requested_contexts = contexts

        # Bind events to the handlers
        evt_handlers = evt_handlers or []
        for evt_hh_args in evt_handlers:
            assoc.bind(*evt_hh_args)

        # Send an A-ASSOCIATE request to the peer and start negotiation
        assoc.request()

        # If the result of the negotiation was acceptance then start up
        #   the Association thread
        if assoc.is_established:
            assoc.start()

        return assoc

    def _create_socket(
        self,
        assoc: Association,
        address: Tuple[str, int],
        tls_args: Optional[Tuple[SSLContext, str]],
    ) -> AssociationSocket:
        """Create an :class:`~pynetdicom.transport.AssociationSocket` for the
        current association.

        .. versionadded:: 1.5
        """
        sock = AssociationSocket(assoc, address=address)
        sock.tls_args = tls_args
        return sock

    @property
    def connection_timeout(self) -> Optional[float]:
        """Get or set the connection timeout (in seconds).

        .. versionadded:: 2.0

        Parameters
        ----------
        value : int, float or None
            The maximum amount of time (in seconds) to wait for a TCP
            connection to be established. A value of ``None`` (default) means
            no timeout. The value is passed to `socket.settimeout()
            <https://docs.python.org/3/library/
            socket.html#socket.socket.settimeout>`_
            and is only used during the connection phase of an association
            request.
        """
        return self._connection_timeout

    @connection_timeout.setter
    def connection_timeout(self, value: Optional[float]) -> None:
        """Set the connection timeout."""
        if value is None:
            self._connection_timeout = None
        # Explicitly excluding zero - this would make the socket non-blocking
        elif isinstance(value, (int, float)) and value > 0:
            self._connection_timeout = value
        else:
            LOGGER.warning("connection_timeout set to None")
            self._connection_timeout = None

        for assoc in self.active_associations:
            assoc.connection_timeout = self.connection_timeout

    @property
    def dimse_timeout(self) -> Optional[float]:
        """Get or set the DIMSE timeout (in seconds).

        Parameters
        ----------
        value : int, float or None
            The maximum amount of time (in seconds) to wait for DIMSE related
            messages. A value of ``None`` means no timeout (default: ``30``).
        """
        return self._dimse_timeout

    @dimse_timeout.setter
    def dimse_timeout(self, value: Optional[float]) -> None:
        """Set the DIMSE timeout in seconds."""
        if value is None:
            self._dimse_timeout = None
        elif isinstance(value, (int, float)) and value >= 0:
            self._dimse_timeout = value
        else:
            LOGGER.warning("dimse_timeout set to 30 s")
            self._dimse_timeout = 30

        for assoc in self.active_associations:
            assoc.dimse_timeout = self.dimse_timeout

    @property
    def implementation_class_uid(self) -> UID:
        """Get or set the *Implementation Class UID* as
        :class:`~pydicom.uid.UID`.

        Parameters
        ----------
        value : str or pydicom.uid.UID
            The association request's *Implementation Class UID* value.
        """
        return self._implementation_uid

    @implementation_class_uid.setter
    def implementation_class_uid(self, value: str) -> None:
        """Set the *Implementation Class UID* used in association requests."""
        uid = cast(UID, set_uid(value, "implementation_class_uid", False, False, True))
        # Enforce conformance on users
        if not uid.is_valid:
            raise ValueError(f"Invalid 'implementation_class_uid' value '{uid}'")

        self._implementation_uid = uid

    @property
    def implementation_version_name(self) -> Optional[str]:
        """Get or set the *Implementation Version Name* as :class:`str`.

        Parameters
        ----------
        value : str or None
            If set then an *Implementation Version Name* item with the
            corresponding value will be added to the association request,
            otherwise no item will be sent.

        Returns
        -------
        str or None
            The set *Implementation Version Name*.
        """
        return self._implementation_version

    @implementation_version_name.setter
    def implementation_version_name(self, value: Optional[str]) -> None:
        """Set the *Implementation Version Name*"""
        # We allow None, but not an empty str
        if isinstance(value, str) and not value:
            raise ValueError(
                "Invalid 'implementation_version_name' value - must not be "
                "an empty str"
            )

        self._implementation_version = set_ae(value, "implementation_version_name")

    def make_server(
        self,
        address: Tuple[str, int],
        ae_title: Optional[str] = None,
        contexts: Optional[ListCXType] = None,
        ssl_context: Optional[SSLContext] = None,
        evt_handlers: Optional[List[EventHandlerType]] = None,
        server_class: Optional[Type[_T]] = None,
        **kwargs: Any,
    ) -> Union[_T, ThreadedAssociationServer]:
        """Return an association server.

        .. versionadded:: 1.5

        Allows the use of a custom association server class.

        Accepts the same parameters as :meth:`start_server`. Additional keyword
        parameters are passed to the constructor of `server_class`.

        .. versionchanged:: 2.0

            `ae_title` should now be :class:`str`

        Parameters
        ----------
        server_class : object, optional
            The class object to use when creating the server. Defaults to
            :class:`~pynetdicom.transport.AssociationServer` if not used.

        Returns
        -------
        object
            The object passed via `server_class` or the
            :class:`~pynetdicom.transport.AssociationServer`.
        """
        # If the SCP has no supported SOP Classes then there's no point
        #   running as a server
        unrestricted = _config.UNRESTRICTED_STORAGE_SERVICE
        if not unrestricted and not contexts and not self.supported_contexts:
            msg = "No supported Presentation Contexts have been defined"
            LOGGER.error(msg)
            raise ValueError(msg)

        ae_title = ae_title if ae_title else self.ae_title
        if isinstance(ae_title, bytes):
            warnings.warn(
                "The use of bytes with 'ae_title' is deprecated, use an "
                "ASCII str instead",
                DeprecationWarning,
            )
            ae_title = decode_bytes(ae_title)

        ae_title = cast(str, set_ae(ae_title, "ae_title", False, False))
        contexts = contexts or self.supported_contexts

        bad_contexts = []
        for cx in contexts:
            roles = (cx.scu_role, cx.scp_role)
            if None in roles and roles != (None, None):
                bad_contexts.append(cx.abstract_syntax)

        if bad_contexts:
            msg = (
                "The following presentation contexts have inconsistent "
                "scu_role/scp_role values (if one is None, both must be):\n  "
            )
            msg += "\n  ".join([str(cx) for cx in bad_contexts])
            raise ValueError(msg)

        server_class = server_class or AssociationServer  # type: ignore[assignment]

        return server_class(  # type: ignore
            self,
            address,
            ae_title,
            contexts,
            ssl_context,
            evt_handlers=evt_handlers or [],
            **kwargs,
        )

    @property
    def maximum_associations(self) -> int:
        """Get or set the number of maximum simultaneous associations as
        :class:`int`.

        Parameters
        ----------
        value : int
            The maximum number of simultaneous associations requested by remote
            AEs. This does not include the number of associations
            requested by the local AE (default ``10``).
        """
        return self._maximum_associations

    @maximum_associations.setter
    def maximum_associations(self, value: int) -> None:
        """Set the number of maximum associations."""
        if isinstance(value, int) and value >= 1:
            self._maximum_associations = value
        else:
            LOGGER.warning("maximum_associations set to 1")
            self._maximum_associations = 1

    @property
    def maximum_pdu_size(self) -> int:
        """Get or set the maximum PDU size accepted by the AE as :class:`int`.

        Parameters
        ----------
        value : int
            The maximum PDU receive size in bytes. A value of ``0`` means the
            PDU size is unlimited (default: ``16382``). Increasing this value
            or setting it to unlimited is an effective way of improving the
            throughput when sending large amounts of data due to the reduced
            DIMSE messaging overhead.
        """
        return self._maximum_pdu_size

    @maximum_pdu_size.setter
    def maximum_pdu_size(self, value: int) -> None:
        """Set the maximum PDU size."""
        # Bounds and type checking of the received maximum length of the
        #   variable field of P-DATA-TF PDUs (in bytes)
        #   * Must be numerical, greater than or equal to 0 (0 indicates
        #       no maximum length (PS3.8 Annex D.1.1)
        if value >= 0:
            self._maximum_pdu_size = value
        else:
            LOGGER.warning(f"maximum_pdu_size set to {DEFAULT_MAX_LENGTH}")

    @property
    def network_timeout(self) -> Optional[float]:
        """Get or set the network timeout (in seconds).

        Parameters
        ----------
        value : int, float or None
            The maximum amount of time (in seconds) to wait for network
            messages. A value of ``None`` means no timeout (default: ``60``).
        """
        return self._network_timeout

    @network_timeout.setter
    def network_timeout(self, value: Optional[float]) -> None:
        """Set the network timeout."""
        if value is None:
            self._network_timeout = None
        elif isinstance(value, (int, float)) and value >= 0:
            self._network_timeout = value
        else:
            LOGGER.warning("network_timeout set to 60 s")
            self._network_timeout = 60

        for assoc in self.active_associations:
            assoc.network_timeout = self.network_timeout

    def remove_requested_context(
        self,
        abstract_syntax: Union[str, UID],
        transfer_syntax: TSyntaxType = None,
    ) -> None:
        """Remove a requested presentation context.

        Depending on the supplied parameters one of the following will occur:

        * `abstract_syntax` alone -  all contexts with a matching abstract
          syntax all be removed.
        * `abstract_syntax` and `transfer_syntax` -  for all contexts with a
          matching abstract syntax; if the supplied `transfer_syntax` list
          contains all of the context's requested transfer syntaxes then the
          entire context will be removed. Otherwise only the matching transfer
          syntaxes will be removed from the context (and the context will
          remain with one or more transfer syntaxes).

        Parameters
        ----------
        abstract_syntax : str, pydicom.uid.UID or sop_class.SOPClass
            The abstract syntax of the presentation context you wish to stop
            requesting when sending association requests.
        transfer_syntax : UID str or list of UID str, optional
            The transfer syntax(es) you wish to stop requesting. If a list of
            str/UID then only those transfer syntaxes specified will no longer
            be requested. If not specified then the abstract syntax and all
            associated transfer syntaxes will no longer be requested (default).

        Examples
        --------
        Remove all requested presentation contexts with an abstract syntax of
        *Verification SOP Class* using its UID value.

        >>> from pynetdicom import AE
        >>> ae = AE()
        >>> ae.add_requested_context('1.2.840.10008.1.1')
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Little Endian
            =Explicit VR Big Endian
        >>> ae.remove_requested_context('1.2.840.10008.1.1')
        >>> len(ae.requested_contexts)
        0

        Remove all requested presentation contexts with an abstract syntax of
        *Verification SOP Class* using the inbuilt
        :class:`~pynetdicom.sop_class.Verification` object.

        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_requested_context(Verification)
        >>> ae.remove_requested_context(Verification)
        >>> len(ae.requested_contexts)
        0

        For all requested presentation contexts with an abstract syntax of
        *Verification SOP Class*, stop requesting a transfer syntax of
        *Implicit VR Little Endian*. If a presentation context exists which
        only has a single *Implicit VR Little Endian* transfer syntax then
        it will be completely removed, otherwise it will be kept with its
        remaining transfer syntaxes.

        Presentation context has only a single matching transfer syntax:

        >>> from pydicom.uid import ImplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae.add_requested_context(Verification, ImplicitVRLittleEndian)
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
        >>> ae.remove_requested_context(Verification, ImplicitVRLittleEndian)
        >>> len(ae.requested_contexts)
        0

        Presentation context has at least one remaining transfer syntax:

        >>> from pydicom.uid import ImplicitVRLittleEndian, ExplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_requested_context(Verification)
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Little Endian
            =Explicit VR Big Endian
        >>> ae.remove_requested_context(
        ...     Verification, [ImplicitVRLittleEndian, ExplicitVRLittleEndian]
        ... )
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Explicit VR Big Endian
        """
        abstract_syntax = UID(abstract_syntax)

        # Get all the current requested contexts with the same abstract syntax
        matching_contexts = [
            cntx
            for cntx in self.requested_contexts
            if cntx.abstract_syntax == abstract_syntax
        ]

        if isinstance(transfer_syntax, str):
            transfer_syntax = [transfer_syntax]

        if transfer_syntax is None:
            # If no transfer_syntax then remove the context completely
            for context in matching_contexts:
                self._requested_contexts.remove(context)
        else:
            for context in matching_contexts:
                for tsyntax in transfer_syntax:
                    if tsyntax in context.transfer_syntax:
                        context.transfer_syntax.remove(UID(tsyntax))

                # Only if all transfer syntaxes have been removed then
                #   remove the context
                if not context.transfer_syntax:
                    self._requested_contexts.remove(context)

    def remove_supported_context(
        self,
        abstract_syntax: Union[str, UID],
        transfer_syntax: TSyntaxType = None,
    ) -> None:
        """Remove a supported presentation context.

        Depending on the supplied parameters one of the following will occur:

        * `abstract_syntax` alone - the entire supported context will be
          removed.
        * `abstract_syntax` and `transfer_syntax` -  If the supplied
          `transfer_syntax` list contains all of the context's supported
          transfer syntaxes then the entire context will be removed.
          Otherwise only the matching transfer syntaxes will be removed from
          the context (and the context will remain with one or more transfer
          syntaxes).

        Parameters
        ----------
        abstract_syntax : str, pydicom.uid.UID or sop_class.SOPClass
            The abstract syntax of the presentation context you wish to stop
            supporting.
        transfer_syntax : UID str or list of UID str, optional
            The transfer syntax(es) you wish to stop supporting. If a list of
            str/UID then only those transfer syntaxes specified will no longer
            be supported. If not specified then the abstract syntax and all
            associated transfer syntaxes will no longer be supported (default).

        Examples
        --------
        Remove the supported presentation context with an abstract syntax of
        *Verification SOP Class* using its UID value.

        >>> from pynetdicom import AE
        >>> ae = AE()
        >>> ae.add_supported_context('1.2.840.10008.1.1')
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Little Endian
            =Explicit VR Big Endian
        >>> ae.remove_supported_context('1.2.840.10008.1.1')
        >>> len(ae.supported_contexts)
        0

        Remove the supported presentation context with an abstract syntax of
        *Verification SOP Class* using the inbuilt
        :class:`~pynetdicom.sop_class.Verification` object.

        >>> from pynetdicom import AE, VerificationPresentationContexts
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.supported_contexts = VerificationPresentationContexts
        >>> ae.remove_supported_context(Verification)

        For the presentation contexts with an abstract syntax of
        *Verification SOP Class*, stop supporting the *Implicit VR Little
        Endian* transfer syntax. If the presentation context only has the
        single *Implicit VR Little Endian* transfer syntax then it will be
        completely removed, otherwise it will be kept with the remaining
        transfer syntaxes.

        Presentation context has only a single matching transfer syntax:

        >>> from pydicom.uid import ImplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_supported_context(Verification, ImplicitVRLittleEndian)
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
        >>> ae.remove_supported_context(Verification, ImplicitVRLittleEndian)
        >>> len(ae.supported_contexts)
        0

        Presentation context has at least one remaining transfer syntax:

        >>> from pydicom.uid import ImplicitVRLittleEndian, ExplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.sop_class import Verification
        >>> ae = AE()
        >>> ae.add_supported_context(Verification)
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian
            =Explicit VR Little Endian
            =Explicit VR Big Endian
        >>> ae.remove_supported_context(
        ...     Verification, [ImplicitVRLittleEndian, ExplicitVRLittleEndian]
        ... )
        >>> print(ae.supported_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Explicit VR Big Endian
        """
        abstract_syntax = UID(abstract_syntax)

        if isinstance(transfer_syntax, str):
            transfer_syntax = [transfer_syntax]

        # Check abstract syntax is actually present
        #   we don't warn if not present because by not being present its not
        #   supported and hence the user's intent has been satisfied
        if abstract_syntax in self._supported_contexts:
            if transfer_syntax is None:
                # If no transfer_syntax then remove the context completely
                del self._supported_contexts[abstract_syntax]
            else:
                # If transfer_syntax then only remove matching syntaxes
                context = self._supported_contexts[abstract_syntax]
                for tsyntax in transfer_syntax:
                    if tsyntax in context.transfer_syntax:
                        context.transfer_syntax.remove(UID(tsyntax))

                # Only if all transfer syntaxes have been removed then remove
                #   the context
                if not context.transfer_syntax:
                    del self._supported_contexts[abstract_syntax]

    @property
    def requested_contexts(self) -> ListCXType:
        """Get or set a list of the requested
        :class:`~pynetdicom.presentation.PresentationContext` items.

        Examples
        --------
        Set the requested presentation contexts using an inbuilt list of
        service specific :class:`~pynetdicom.presentation.PresentationContext`
        items:

        >>> from pynetdicom import AE, StoragePresentationContexts
        >>> ae = AE()
        >>> ae.requested_contexts = StoragePresentationContexts

        Set the requested presentation contexts using a :class:`list` of
        :class:`~pynetdicom.presentation.PresentationContext` items:

        >>> from pydicom.uid import ImplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.presentation import PresentationContext
        >>> context = PresentationContext()
        >>> context.abstract_syntax = '1.2.840.10008.1.1'
        >>> context.transfer_syntax = [ImplicitVRLittleEndian]
        >>> ae = AE()
        >>> ae.requested_contexts = [context]
        >>> print(ae.requested_contexts[0])
        Abstract Syntax: Verification SOP Class
        Transfer Syntax(es):
            =Implicit VR Little Endian

        Parameters
        ----------
        contexts : list of PresentationContext
            The presentation contexts to request when acting as an SCU.

        Raises
        ------
        ValueError
            If trying to add more than 128 requested presentation contexts.

        See Also
        --------
        ApplicationEntity.add_requested_context
            Add a single presentation context to the requested contexts using
            an abstract syntax and (optionally) a list of transfer syntaxes.
        """
        return self._requested_contexts

    @requested_contexts.setter
    def requested_contexts(self, contexts: ListCXType) -> None:
        """Set the requested presentation contexts."""
        if not contexts:
            self._requested_contexts = []
            return

        self._validate_requested_contexts(contexts)

        for context in contexts:
            self.add_requested_context(
                cast(UID, context.abstract_syntax), context.transfer_syntax
            )

    @property
    def require_called_aet(self) -> bool:
        """Get or set whether the *Called AE Title* must match the AE title.

        When an association request is received the value of the 'Called AE
        Title' supplied by the peer will be compared with the set values and
        if none match the association will be rejected. If the set value
        is an empty list then the *Called AE Title* will not be checked.

        .. versionchanged:: 1.1

            `require_match` changed to ``bool``

        Parameters
        ----------
        require_match : bool
            If ``True`` then any association requests that supply a
            *Called AE Title* value that does not match :attr:`ae_title`
            will be rejected. If ``False`` (default) then all association
            requests will be accepted (unless rejected for other reasons).
        """
        return self._require_called_aet

    @require_called_aet.setter
    def require_called_aet(self, require_match: bool) -> None:
        """Set whether the *Called AE Title* must match the AE title."""
        self._require_called_aet = require_match

    @property
    def require_calling_aet(self) -> List[str]:
        """Get or set the required calling AE title as a list of :class:`str`.

        When an association request is received the value of the *Calling AE
        Title* supplied by the peer will be compared with the set value and
        if none match the association will be rejected. If the set value
        is an empty list then the *Calling AE Title* will not be checked.

        .. versionchanged:: 1.1

            `ae_titles` changed to :class:`list` of :class:`bytes`

        .. versionchanged:: 2.0

            `ae_titles` should now be a :class:`list` of :class:`str`

        Parameters
        ----------
        ae_titles : list of str
            If not empty then any association requests that supply a
            *Calling AE Title* value that does not match one of the values in
            *ae_titles* will be rejected. If an empty list (default) then all
            association requests will be accepted (unless rejected for other
            reasons).
        """
        return self._require_calling_aet

    @require_calling_aet.setter
    def require_calling_aet(self, ae_titles: List[str]) -> None:
        """Set the required calling AE title."""
        if any([isinstance(v, bytes) for v in ae_titles]):
            warnings.warn(
                "The use of a list of bytes with 'require_calling_aet' is "
                "deprecated, use a list of ASCII str instead",
                DeprecationWarning,
            )

        values = []
        for v in ae_titles:
            if isinstance(v, bytes):
                v = decode_bytes(v)

            values.append(cast(str, set_ae(v, "require_calling_aet", False, False)))

        self._require_calling_aet = values

    def shutdown(self) -> None:
        """Stop any active association servers and threads.

        .. versionadded:: 1.2
        """
        for assoc in self.active_associations:
            assoc.abort()

        # This is a bit hackish: server.shutdown() removes the server
        #   from `_servers` so we need to workaround this
        for server in self._servers[:]:
            server.shutdown()

        self._servers = []

    def start_server(
        self,
        address: Tuple[str, int],
        block: bool = True,
        ssl_context: Optional[SSLContext] = None,
        evt_handlers: Optional[List[EventHandlerType]] = None,
        ae_title: Optional[str] = None,
        contexts: Optional[ListCXType] = None,
    ) -> Optional[ThreadedAssociationServer]:
        """Start the AE as an association *acceptor*.

        .. versionadded:: 1.2

        If set to non-blocking then a running
        :class:`~pynetdicom.transport.ThreadedAssociationServer`
        instance will be returned. This can be stopped using
        :meth:`~pynetdicom.transport.AssociationServer.shutdown`.

        .. versionchanged:: 1.3

            Added `evt_handlers` keyword parameter

        .. versionchanged:: 1.4

            Added `ae_title` and `contexts` keyword parameters

        .. versionchanged:: 1.5

            `evt_handlers` now takes a list of 2- or 3-tuples

        .. versionchanged:: 2.0

            `ae_title` should now be :class:`str`

        Parameters
        ----------
        address : Tuple[str, int]
            The ``(host: str, port: int)`` to use when listening for incoming
            association requests.
        block : bool, optional
            If ``True`` (default) then the server will be blocking, otherwise
            it will start the server in a new thread and be non-blocking.
        ssl_context : ssl.SSLContext, optional
            If TLS is required then this should the :class:`ssl.SSLContext`
            instance to use to wrap the client sockets, otherwise if ``None``
            then no TLS will be used (default).
        evt_handlers : list of 2- or 3-tuple, optional
            A list of (*event*, *handler*) or (*event*, *handler*, *args*),
            where `event` is an ``evt.EVT_*`` event tuple, `handler` is a
            callable function that will be bound to the event and `args` is a
            :class:`list` of objects that will be passed to `handler` as
            optional extra arguments. At a minimum, `handler` should take an
            :class:`~pynetdicom.events.Event` parameter and may return or yield
            objects depending on the exact event that the handler is bound to.
            For more information see the :ref:`documentation<user_events>`.
        ae_title : str, optional
            The AE title to use for the local SCP. If this keyword parameter
            is not used then the AE title from the :attr:`ae_title` property
            will be used instead (default).
        contexts : list of presentation.PresentationContext, optional
            The presentation contexts that will be supported by the SCP. If
            not used then the presentation contexts in the
            :attr:`supported_contexts` property will be used instead (default).

        Returns
        -------
        transport.ThreadedAssociationServer or None
            If `block` is ``False`` then returns the server instance, otherwise
            returns ``None``.
        """
        if block:
            # Blocking server
            server = self.make_server(
                address,
                ae_title=ae_title,
                contexts=contexts,
                ssl_context=ssl_context,
                evt_handlers=evt_handlers,
            )
            self._servers.append(server)

            try:
                # **BLOCKING**
                server.serve_forever()
            except KeyboardInterrupt:
                server.shutdown()

            return None

        # Non-blocking server
        timestamp = datetime.strftime(datetime.now(), "%Y%m%d%H%M%S")
        server = self.make_server(
            address,
            ae_title=ae_title,
            contexts=contexts,
            ssl_context=ssl_context,
            evt_handlers=evt_handlers,
            server_class=ThreadedAssociationServer,
        )

        thread = threading.Thread(
            target=make_target(server.serve_forever), name=f"AcceptorServer@{timestamp}"
        )
        thread.daemon = True
        thread.start()

        self._servers.append(server)

        return server

    def __str__(self) -> str:
        """Prints out the attribute values and status for the AE"""
        s = [""]
        s.append(f"Application Entity {self.ae_title}")

        s.append("")
        s.append("  Requested Presentation Contexts:")
        if not self.requested_contexts:
            s.append("\tNone")
        for context in self.requested_contexts:
            s.append(f"\t{cast(UID, context.abstract_syntax).name}")
            for transfer_syntax in context.transfer_syntax:
                s.append(f"\t\t{transfer_syntax.name}")

        s.append("")
        s.append("  Supported Presentation Contexts:")
        if not self.supported_contexts:
            s.append("\tNone")
        for context in self.supported_contexts:
            s.append(f"\t{cast(UID, context.abstract_syntax).name}")
            for transfer_syntax in context.transfer_syntax:
                s.append(f"\t\t{transfer_syntax.name}")

        s.append("")
        s.append(f"  ACSE timeout: {self.acse_timeout} s")
        s.append(f"  DIMSE timeout: {self.dimse_timeout} s")
        s.append(f"  Network timeout: {self.network_timeout} s")
        s.append(f"  Connection timeout: {self.connection_timeout} s")

        s.append("")
        if self.require_calling_aet != []:
            ae_titles = self.require_calling_aet
            s.append((f"  Required calling AE title(s): {', '.join(ae_titles)}"))
        s.append(f"  Require called AE title: {self.require_called_aet}")
        s.append("")

        # Association information
        s.append(
            f"  Association(s): {len(self.active_associations)}"
            f"/{self.maximum_associations}"
        )

        for assoc in self.active_associations:
            s.append(
                f"\tPeer: {assoc.remote['ae_title']} on "
                f"{assoc.remote['address']}:{assoc.remote['port']}"
            )

        return "\n".join(s)

    @property
    def supported_contexts(self) -> ListCXType:
        """Get or set a list of the supported
        :class:`~pynetdicom.presentation.PresentationContext` items.

        Examples
        --------
        Set the supported presentation contexts using a list of
        ``PresentationContext`` items:

        >>> from pydicom.uid import ImplicitVRLittleEndian
        >>> from pynetdicom import AE
        >>> from pynetdicom.presentation import PresentationContext
        >>> context = PresentationContext()
        >>> context.abstract_syntax = '1.2.840.10008.1.1'
        >>> context.transfer_syntax = [ImplicitVRLittleEndian]
        >>> ae = AE()
        >>> ae.supported_contexts = [context]

        Set the supported presentation contexts using an inbuilt list of
        service specific :class:`~pynetdicom.presentation.PresentationContext`
        items:

        >>> from pynetdicom import AE, StoragePresentationContexts
        >>> ae = AE()
        >>> ae.supported_contexts = StoragePresentationContexts

        Parameters
        ----------
        contexts : list of presentation.PresentationContext
            The presentation contexts to support when acting as an SCP.

        See Also
        --------
        ApplicationEntity.add_supported_context
            Add a single presentation context to the supported contexts using
            an abstract syntax and optionally a list of transfer syntaxes.
        """
        # The supported presentation contexts are stored internally as a dict
        return sorted(
            list(self._supported_contexts.values()),
            key=lambda cx: cast(UID, cx.abstract_syntax),
        )

    @supported_contexts.setter
    def supported_contexts(self, contexts: ListCXType) -> None:
        """Set the supported presentation contexts using a list."""
        if not contexts:
            self._supported_contexts = {}

        for item in contexts:
            if not isinstance(item, PresentationContext):
                raise ValueError(
                    "'contexts' must be a list of PresentationContext items"
                )

            self.add_supported_context(
                cast(UID, item.abstract_syntax), item.transfer_syntax
            )

    @staticmethod
    def _validate_requested_contexts(contexts: ListCXType) -> None:
        """Validate the supplied `contexts`.

        Parameters
        ----------
        contexts : list of presentation.PresentationContext
            The contexts to validate.
        """
        if len(contexts) > 128:
            raise ValueError(
                "The maximum allowed number of requested presentation "
                "contexts is 128"
            )

        invalid = [ii for ii in contexts if not isinstance(ii, PresentationContext)]
        if invalid:
            raise ValueError("'contexts' must be a list of PresentationContext items")



"""Unit tests for the events module."""

from collections import namedtuple
from datetime import datetime
from io import BytesIO
import logging
import os
import sys
import time

import pytest

from pydicom.dataset import Dataset
from pydicom.tag import BaseTag
from pydicom.uid import ImplicitVRLittleEndian
from pydicom.filewriter import write_file_meta_info

from pynetdicom import (
    AE,
    evt,
    _config,
    Association,
    debug_logger,
    build_context,
    PYNETDICOM_IMPLEMENTATION_UID,
    PYNETDICOM_IMPLEMENTATION_VERSION,
)
from pynetdicom.events import (
    Event,
    trigger,
    _async_ops_handler,
    _sop_common_handler,
    _sop_extended_handler,
    _user_identity_handler,
    _c_echo_handler,
    _c_get_handler,
    _c_find_handler,
    _c_move_handler,
    _c_store_handler,
    _n_action_handler,
    _n_create_handler,
    _n_delete_handler,
    _n_event_report_handler,
    _n_get_handler,
    _n_set_handler,
)
from pynetdicom.dimse_messages import (
    N_ACTION,
    N_CREATE,
    N_EVENT_REPORT,
    N_SET,
    N_GET,
    N_DELETE,
    C_STORE,
)
from pynetdicom.sop_class import Verification


# debug_logger()


def test_intervention_namedtuple():
    """Test the InterventionEvent namedtuple."""
    event = evt.InterventionEvent("some name", "some description")
    assert event.name == "some name"
    assert event.description == "some description"
    assert event.is_intervention is True
    assert event.is_notification is False
    assert str(event) == event.name


def test_notification_namedtuple():
    """Test the NotificationEvent namedtuple."""
    event = evt.NotificationEvent("some name", "some description")
    assert event.name == "some name"
    assert event.description == "some description"
    assert event.is_intervention is False
    assert event.is_notification is True
    assert str(event) == event.name


def test_intervention_global():
    """Test the _INTERVENTION_EVENTS global."""
    assert evt.EVT_C_ECHO in evt._INTERVENTION_EVENTS
    assert evt.EVT_DATA_RECV not in evt._INTERVENTION_EVENTS


def test_notification_global():
    """Test the _NOTIFICATION_EVENTS global."""
    assert evt.EVT_C_ECHO not in evt._NOTIFICATION_EVENTS
    assert evt.EVT_DATA_RECV in evt._NOTIFICATION_EVENTS


class TestEvent:
    """Tests for event.Event."""

    def setup_method(self):
        self.ae = None
        _config.LOG_HANDLER_LEVEL = "none"

        # Implicit VR Little Endian
        self.bytestream = BytesIO(
            #  (0010,0010) PatientName
            # | tag           | length        | value
            b"\x10\x00\x10\x00\x09\x00\x00\x00"
            b"\x54\x45\x53\x54\x5E\x54\x65\x73\x74"
        )
        self.context = build_context("1.2.840.10008.1.1", ImplicitVRLittleEndian)

    def teardown_method(self):
        if self.ae:
            self.ae.shutdown()

        _config.LOG_HANDLER_LEVEL = "standard"

    def test_init(self):
        """Test initialisation of event.Event."""
        event = evt.Event(None, evt.EVT_C_STORE)
        assert event.assoc is None
        assert event._event == evt.EVT_C_STORE
        assert isinstance(event.timestamp, datetime)
        assert event.event == evt.EVT_C_STORE
        assert event.event.name == "EVT_C_STORE"
        assert isinstance(event.event.description, str)

        def callable():
            return "some value"

        event = evt.Event(
            None, evt.EVT_C_STORE, {"aa": True, "bb": False, "cc": callable}
        )
        assert event.assoc is None
        assert event._event == evt.EVT_C_STORE
        assert event.event == evt.EVT_C_STORE
        assert isinstance(event.timestamp, datetime)
        assert event.event.name == "EVT_C_STORE"
        assert isinstance(event.event.description, str)
        assert event.cc() == "some value"
        assert event.aa is True
        assert event.bb is False

    def test_raises(self):
        """Test property getters raise if not correct event type."""
        event = evt.Event(None, evt.EVT_DATA_RECV)
        msg = (
            r"The corresponding event is not a C-STORE "
            r"request and has no 'Data Set' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.dataset

        msg = (
            r"The corresponding event is not a C-STORE request "
            r"and has no 'Data Set' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.encoded_dataset()

        msg = (
            r"The corresponding event is not a C-FIND, C-GET or C-MOVE request "
            r"and has no 'Identifier' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.identifier

        msg = (
            r"The corresponding event is not a C-MOVE request "
            r"and has no 'Move Destination' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.move_destination

        msg = (
            r"The corresponding event is not an N-ACTION request and has no "
            r"'Action Information' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.action_information

        msg = (
            r"The corresponding event is not an N-CREATE request and has no "
            r"'Attribute List' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.attribute_list

        msg = (
            r"The corresponding event is not an N-EVENT-REPORT request and "
            r"has no 'Event Information' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.event_information

        msg = (
            r"The corresponding event is not an N-SET request and has no "
            r"'Modification List' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.modification_list

        msg = (
            r"The corresponding event is not an N-GET request and has no "
            r"'Attribute Identifier List' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.attribute_identifiers

        msg = (
            r"The corresponding event is not an N-ACTION request and has no "
            r"'Action Type ID' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.action_type

        msg = (
            r"The corresponding event is not an N-EVENT-REPORT request and "
            r"has no 'Event Type ID' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.event_type

        msg = (
            r"The corresponding event is not a DIMSE service request and "
            r"has no 'Message ID' parameter"
        )
        with pytest.raises(AttributeError, match=msg):
            event.message_id

        msg = (
            r"The corresponding event is either not a C-STORE request or "
            r"'STORE_RECV_CHUNKED_DATASET' is not True."
        )
        with pytest.raises(AttributeError, match=msg):
            event.dataset_path

    def test_is_cancelled_non(self):
        """Test Event.is_cancelled with wrong event type."""
        event = evt.Event(None, evt.EVT_DATA_RECV)
        assert event.is_cancelled is False

        def is_cancelled(msg_id):
            if msg_id in [1, 2, 3]:
                return True

            return False

        # No Message ID
        event = evt.Event(None, evt.EVT_DATA_RECV, {"_is_cancelled": is_cancelled})
        assert event.is_cancelled is False

    def test_is_cancelled(self):
        """Test Event.is_cancelled with correct event type."""

        def is_cancelled(msg_id):
            if msg_id in [1, 2, 3]:
                return True

            return False

        message = namedtuple("Message", ["MessageID"])
        msg = message(1)
        msg2 = message(7)

        event = evt.Event(
            None, evt.EVT_DATA_RECV, {"_is_cancelled": is_cancelled, "request": msg}
        )
        assert event.is_cancelled is True

        event = evt.Event(
            None, evt.EVT_DATA_RECV, {"_is_cancelled": is_cancelled, "request": msg2}
        )
        assert event.is_cancelled is False

    def test_assign_existing(self):
        """Test adding an attribute that already exists."""
        msg = r"'Event' object already has an attribute 'assoc'"
        with pytest.raises(AttributeError, match=msg):
            event = evt.Event(None, evt.EVT_C_STORE, {"assoc": None})

    def test_action_information(self):
        """Test Event.action_information."""
        request = N_ACTION()
        request.ActionInformation = self.bytestream
        event = Event(
            None,
            evt.EVT_N_ACTION,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event._hash is None
        assert event._decoded is None
        ds = event.action_information
        assert event._hash == hash(request.ActionInformation)
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "TEST^Test"

        ds.PatientID = "1234567"
        assert event.action_information.PatientID == "1234567"

        # Test hash mismatch
        event._hash = None
        assert "PatientID" not in event.action_information

    def test_attribute_list(self):
        """Test Event.attribute_list."""
        request = N_CREATE()
        request.AttributeList = self.bytestream
        event = Event(
            None,
            evt.EVT_N_CREATE,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event._hash is None
        assert event._decoded is None
        ds = event.attribute_list
        assert event._hash == hash(request.AttributeList)
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "TEST^Test"

        ds.PatientID = "1234567"
        assert event.attribute_list.PatientID == "1234567"

        # Test hash mismatch
        event._hash = None
        assert "PatientID" not in event.attribute_list

    def test_event_information(self):
        """Test Event.event_information."""
        request = N_EVENT_REPORT()
        request.EventInformation = self.bytestream
        event = Event(
            None,
            evt.EVT_N_CREATE,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event._hash is None
        assert event._decoded is None
        ds = event.event_information
        assert event._hash == hash(request.EventInformation)
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "TEST^Test"

        ds.PatientID = "1234567"
        assert event.event_information.PatientID == "1234567"

        # Test hash mismatch
        event._hash = None
        assert "PatientID" not in event.event_information

    def test_file_meta(self):
        """Test Event.file_meta."""
        request = C_STORE()
        request.AffectedSOPClassUID = "1.2.3.4"
        request.AffectedSOPInstanceUID = "1.2.3.4.5"
        request.DataSet = BytesIO(b"\x00\x01")

        event = Event(
            None,
            evt.EVT_C_STORE,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event._hash is None
        assert event._decoded is None
        meta = event.file_meta
        assert event._hash is None
        assert event._decoded is None
        assert 0 == meta.FileMetaInformationGroupLength
        assert b"\x00\x01" == meta.FileMetaInformationVersion
        assert "1.2.3.4" == meta.MediaStorageSOPClassUID
        assert "1.2.3.4.5" == meta.MediaStorageSOPInstanceUID
        assert ImplicitVRLittleEndian == meta.TransferSyntaxUID
        assert PYNETDICOM_IMPLEMENTATION_UID == meta.ImplementationClassUID
        assert PYNETDICOM_IMPLEMENTATION_VERSION == (meta.ImplementationVersionName)

    def test_write_file_meta(self):
        """pydicom-independent test confirming correct write."""
        request = C_STORE()
        request.AffectedSOPClassUID = "1.2"
        request.AffectedSOPInstanceUID = "1.3"
        request.DataSet = BytesIO(b"\x00\x01")

        event = Event(
            None,
            evt.EVT_C_STORE,
            {"request": request, "context": self.context.as_tuple},
        )
        fp = BytesIO()
        meta = event.file_meta
        write_file_meta_info(fp, meta)
        bs = fp.getvalue()
        assert bs[:12] == b"\x02\x00\x00\x00\x55\x4c\x04\x00\x7e\x00\x00\x00"
        assert bs[12:76] == (
            b"\x02\x00\x01\x00\x4f\x42\x00\x00\x02\x00\x00\x00\x00\x01"
            b"\x02\x00\x02\x00\x55\x49\x04\x00\x31\x2e\x32\x00"
            b"\x02\x00\x03\x00\x55\x49\x04\x00\x31\x2e\x33\x00"
            b"\x02\x00\x10\x00\x55\x49\x12\x00\x31\x2e\x32\x2e\x38\x34"
            b"\x30\x2e\x31\x30\x30\x30\x38\x2e\x31\x2e\x32\x00"
        )

        # Note: may not be 126 if Implementation Class and Version change
        assert 126 == meta.FileMetaInformationGroupLength
        assert 12 + 126 == len(fp.getvalue())

    def test_modification_list(self):
        """Test Event.modification_list."""
        request = N_SET()
        request.ModificationList = self.bytestream
        event = Event(
            None,
            evt.EVT_N_CREATE,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event._hash is None
        assert event._decoded is None
        ds = event.modification_list
        assert event._hash == hash(request.ModificationList)
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "TEST^Test"

        ds.PatientID = "1234567"
        assert event.modification_list.PatientID == "1234567"

        # Test hash mismatch
        event._hash = None
        assert "PatientID" not in event.modification_list

    def test_message_id(self):
        """Test Event.modification_list."""
        request = N_SET()
        request.MessageID = 1234
        event = Event(
            None,
            evt.EVT_N_CREATE,
            {"request": request, "context": self.context.as_tuple},
        )

        assert 1234 == event.message_id

    def test_empty_dataset(self):
        """Test with an empty dataset-like."""
        request = N_EVENT_REPORT()
        event = Event(
            None,
            evt.EVT_N_CREATE,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event._hash is None
        assert event._decoded is None
        ds = event.event_information
        assert event._hash == hash(request.EventInformation)
        assert ds == Dataset()

        # Test in-place modification works OK
        ds.PatientID = "1234567"
        assert event.event_information.PatientID == "1234567"

        # Test hash mismatch
        event._hash = None
        assert "PatientID" not in event.event_information

    def test_empty_attr_identifiers(self):
        """Test with an empty attribute_identifiers."""
        request = N_GET()
        event = Event(
            None, evt.EVT_N_GET, {"request": request, "context": self.context.as_tuple}
        )

        assert event.attribute_identifiers == []

    def test_attr_identifiers(self):
        """Test with attribute_identifiers."""
        request = N_GET()
        request.AttributeIdentifierList = [0x00100010, 0x00100020]
        event = Event(
            None, evt.EVT_N_GET, {"request": request, "context": self.context.as_tuple}
        )

        tags = event.attribute_identifiers
        assert isinstance(tags[0], BaseTag)
        assert tags[0] == 0x00100010
        assert isinstance(tags[1], BaseTag)
        assert tags[1] == 0x00100020

        request.AttributeIdentifierList = 0x00100010
        event = Event(
            None, evt.EVT_N_GET, {"request": request, "context": self.context.as_tuple}
        )

        tags = event.attribute_identifiers
        assert tags == [0x00100010]

    def test_action_type(self):
        """Test with action_type."""
        request = N_ACTION()
        request.ActionTypeID = 2
        event = Event(
            None,
            evt.EVT_N_ACTION,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event.action_type == 2

    def test_event_type(self):
        """Test with event_type."""
        request = N_EVENT_REPORT()
        request.EventTypeID = 2
        event = Event(
            None,
            evt.EVT_N_EVENT_REPORT,
            {"request": request, "context": self.context.as_tuple},
        )

        assert event.event_type == 2

    def test_encode_dataset(self):
        """Test Event.encode_dataset()"""
        request = C_STORE()
        request.AffectedSOPClassUID = "1.2"
        request.AffectedSOPInstanceUID = "1.3"
        request.DataSet = BytesIO(b"\x00\x01")

        event = Event(
            None,
            evt.EVT_C_STORE,
            {"request": request, "context": self.context.as_tuple},
        )
        bs = event.encoded_dataset()

        from pynetdicom.utils import pretty_bytes

        assert bs[:128] == b"\x00" * 128
        assert bs[128:132] == b"DICM"
        assert bs[132:144] == b"\x02\x00\x00\x00\x55\x4c\x04\x00\x7e\x00\x00\x00"
        assert bs[144:144 + 64] == (
            b"\x02\x00\x01\x00\x4f\x42\x00\x00\x02\x00\x00\x00\x00\x01"
            b"\x02\x00\x02\x00\x55\x49\x04\x00\x31\x2e\x32\x00"
            b"\x02\x00\x03\x00\x55\x49\x04\x00\x31\x2e\x33\x00"
            b"\x02\x00\x10\x00\x55\x49\x12\x00\x31\x2e\x32\x2e\x38\x34"
            b"\x30\x2e\x31\x30\x30\x30\x38\x2e\x31\x2e\x32\x00"
        )

        # Note: may not be 126 if Implementation Class and Version change
        assert 128 + 4 + 12 + 126 + 2== len(bs)

        # Test without file_meta
        bs = event.encoded_dataset(include_meta=False)
        assert bs == b"\x00\x01"


# TODO: Should be able to remove in v1.4
INTERVENTION_HANDLERS = [
    _async_ops_handler,
    _sop_common_handler,
    _sop_extended_handler,
    _user_identity_handler,
    _c_echo_handler,
    _c_get_handler,
    _c_find_handler,
    _c_move_handler,
    _c_store_handler,
    _n_action_handler,
    _n_create_handler,
    _n_delete_handler,
    _n_event_report_handler,
    _n_get_handler,
    _n_set_handler,
]


@pytest.mark.parametrize("handler", INTERVENTION_HANDLERS)
def test_default_handlers(handler):
    if handler not in [_sop_common_handler, _sop_extended_handler, _c_echo_handler]:
        with pytest.raises(NotImplementedError):
            handler(None)
    else:
        handler(None)



"""Tests for the service_class module."""

import queue

import pytest

from pydicom.dataset import Dataset

from pynetdicom import build_context
from pynetdicom.dimse_primitives import C_STORE, C_GET, C_MOVE, C_CANCEL
from pynetdicom.service_class import StorageServiceClass, ServiceClass


class DummyAssoc:
    def __init__(self):
        self.dimse = DummyDIMSE()


class DummyDIMSE:
    def __init__(self):
        self.msg_queue = queue.Queue()
        self.dimse_timeout = 0.5
        self.cancel_req = {}

    def get_msg(self, block=True):
        try:
            return self.msg_queue.get(block=block)
        except queue.Empty:
            return (None, None)

    def peek_msg(self):
        try:
            return self.msg_queue.queue[0]
        except (queue.Empty, IndexError):
            return (None, None)


class TestServiceClass:
    def test_is_valid_status(self):
        """Test that is_valid_status returns correct values"""
        sop = StorageServiceClass(None)
        assert not sop.is_valid_status(0x0101)
        assert sop.is_valid_status(0x0000)

    def test_validate_status_ds(self):
        """Test that validate_status works correctly with dataset"""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        status = Dataset()
        status.Status = 0x0001
        rsp = sop.validate_status(status, rsp)
        assert rsp.Status == 0x0001

    def test_validate_status_ds_multi(self):
        """Test that validate_status works correctly with dataset multi"""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        status = Dataset()
        status.Status = 0x0002
        status.ErrorComment = "test"
        rsp = sop.validate_status(status, rsp)
        assert rsp.Status == 0x0002
        assert rsp.ErrorComment == "test"

    def test_validate_status_ds_none(self):
        """Test correct status returned if ds has no Status element."""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        status = Dataset()
        status.ErrorComment = "Test comment"
        rsp = sop.validate_status(status, rsp)
        assert rsp.Status == 0xC001

    def test_validate_status_ds_unknown(self):
        """Test a status ds with an unknown element."""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        status = Dataset()
        status.Status = 0x0000
        status.PatientName = "Test comment"
        rsp = sop.validate_status(status, rsp)

    def test_validate_status_int(self):
        """Test that validate_status works correctly with int"""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        rsp = sop.validate_status(0x0000, rsp)
        assert rsp.Status == 0x0000

    def test_validate_status_invalid(self):
        """Test exception raised if invalid status value"""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        rsp = sop.validate_status("test", rsp)
        assert rsp.Status == 0xC002

    def test_validate_status_unknown(self):
        """Test return unknown status"""
        sop = StorageServiceClass(None)
        rsp = C_STORE()
        rsp = sop.validate_status(0xD011, rsp)
        assert rsp.Status == 0xD011

    def test_scp_raises(self):
        """Test that ServiceClass.SCP raises exception"""
        service = ServiceClass(None)
        msg = r"No service class has been implemented for the " r"SOP Class UID '1.2.3'"
        with pytest.raises(NotImplementedError, match=msg):
            service.SCP(None, build_context("1.2.3"))

    def test_is_cancelled_no_msg(self):
        """Test is_cancelled with no DIMSE messages in the queue."""
        assoc = DummyAssoc()
        service = ServiceClass(assoc)
        assert service.is_cancelled(1) is False
        assert service.is_cancelled(2) is False

    def test_is_cancelled_no_match(self):
        """Test is_cancelled with no matching C-CANCEL."""
        assoc = DummyAssoc()
        cancel = C_CANCEL()
        cancel.MessageIDBeingRespondedTo = 5
        assoc.dimse.cancel_req[5] = cancel
        assoc.dimse.cancel_req[3] = cancel
        service = ServiceClass(assoc)
        assert service.is_cancelled(1) is False
        assert service.is_cancelled(2) is False

    def test_is_cancelled_match(self):
        """Test is_cancelled with matching C-CANCEL."""
        assoc = DummyAssoc()
        cancel = C_CANCEL()
        cancel.MessageIDBeingRespondedTo = 5
        assoc.dimse.cancel_req[4] = C_GET()
        assoc.dimse.cancel_req[3] = cancel
        service = ServiceClass(assoc)
        assert service.is_cancelled(1) is False
        assert service.is_cancelled(2) is False
        assert service.is_cancelled(3) is True
        service = ServiceClass(assoc)
        assert service.is_cancelled(1) is False
        assert service.is_cancelled(2) is False
        assert service.is_cancelled(3) is False
        assert cancel not in assoc.dimse.cancel_req.values()



"""Unit tests for the pynetdicom.utils module."""

from codecs import BOM_UTF32_LE
from io import BytesIO
from threading import Thread
import logging
import sys

import pytest

from pydicom.uid import UID

from pynetdicom import _config, debug_logger
from pynetdicom.utils import (
    pretty_bytes,
    validate_uid,
    make_target,
    set_uid,
    decode_bytes,
)
from .encoded_pdu_items import a_associate_rq


# debug_logger()


REFERENCE_GOOD_AE_STR = [
    ("a", b"a               "),
    ("a              b", b"a              b"),
    ("a    b", b"a    b          "),
    ("               b", b"b               "),
    ("        ab  c   ", b"ab  c           "),
    ("        ab  c      ", b"ab  c           "),
    ("ABCDEFGHIJKLMNOPQRSTUVWXYZ", b"ABCDEFGHIJKLMNOP"),
]
REFERENCE_GOOD_AE_BYTES = [
    (b"a", b"a               "),
    (b"a              b", b"a              b"),
    (b"a    b", b"a    b          "),
    (b"               b", b"b               "),
    (b"        ab  c   ", b"ab  c           "),
    (b"        ab  c      ", b"ab  c           "),
    (b"ABCDEFGHIJKLMNOPQRSTUVWXYZ", b"ABCDEFGHIJKLMNOP"),
]
REFERENCE_BAD_AE_STR = [
    "                ",  # empty, 16 chars 0x20
    "",  # empty
    "AE\\TITLE",  # backslash
    "AE\tTITLE",  # control char, tab
    "AE\rTITLE",  # control char, carriage return
    "AE\nTITLE",  # control char, new line
    "\u0009".encode("ascii"),  # \t
    "\u000A".encode("ascii"),  # \n
    "\u000C".encode("ascii"),  # \x0c
    "\u000D".encode("ascii"),  # \x0d
    "\u001B".encode("ascii"),  # \x1b
    "\u005C".encode("ascii"),  # \\
    "\u0001".encode("ascii"),  # \x01
    "\u000e".encode("ascii"),  # \x0e
    1234,
    45.1,
]
REFERENCE_BAD_AE_BYTES = [
    b"                ",  # empty, 16 chars 0x20
    b"",  # empty
    b"AE\\TITLE",  # backslash
    b"AE\tTITLE",  # control char, tab
    b"AE\rTITLE",  # control char, carriage return
    b"AE\nTITLE",  # control char, new line
    "\u0009".encode("ascii"),  # \t
    "\u000A".encode("ascii"),  # \n
    "\u000C".encode("ascii"),  # \x0c
    "\u000D".encode("ascii"),  # \x0d
    "\u001B".encode("ascii"),  # \x1b
    "\u005C".encode("ascii"),  # \\
    "\u0001".encode("ascii"),  # \x01
    "\u000e".encode("ascii"),  # \x0e
    1234,
    45.1,
]


REFERENCE_UID = [
    # UID, (enforced, non-enforced conformance)
    # Invalid, invalid
    ("", (False, False)),
    (" " * 64, (False, False)),
    ("1" * 65, (False, False)),
    ("a" * 65, (False, False)),
    # Invalid, valid
    ("a" * 64, (False, True)),
    ("0.1.2.04", (False, True)),
    ("some random string", (False, True)),
    # Valid, valid
    ("1" * 64, (True, True)),
    ("0.1.2.4", (True, True)),
]


class TestValidateUID:
    """Tests for utils.validate_uid()."""

    def setup_method(self):
        self.default_conformance = _config.ENFORCE_UID_CONFORMANCE

    def teardown_method(self):
        _config.ENFORCE_UID_CONFORMANCE = self.default_conformance

    @pytest.mark.parametrize("uid,is_valid", REFERENCE_UID)
    def test_validate_uid_conformance_true(self, uid, is_valid):
        _config.ENFORCE_UID_CONFORMANCE = True
        assert validate_uid(UID(uid)) == is_valid[0]

    @pytest.mark.parametrize("uid,is_valid", REFERENCE_UID)
    def test_validate_uid_conformance_false(self, uid, is_valid):
        _config.ENFORCE_UID_CONFORMANCE = False
        assert validate_uid(UID(uid)) == is_valid[1]


class TestPrettyBytes:
    """Tests for utils.pretty_bytes()."""

    def test_parameters(self):
        """Test parameters are correct."""
        # Default
        bytestream = a_associate_rq
        result = pretty_bytes(bytestream)
        assert len(result) == 14
        assert isinstance(result[0], str)

        # prefix
        result = pretty_bytes(bytestream, prefix="\\x")
        for line in result:
            assert line[:2] == "\\x"

        # delimiter
        result = pretty_bytes(bytestream, prefix="", delimiter=",")
        for line in result:
            assert line[2] == ","

        # items_per_line
        result = pretty_bytes(bytestream, prefix="", delimiter="", items_per_line=10)
        assert len(result[0]) == 20

        # max_size
        result = pretty_bytes(
            bytestream, prefix="", delimiter="", items_per_line=10, max_size=100
        )
        assert len(result) == 11  # 10 plus the cutoff line
        assert "Only dumping 100 bytes." == result[0]
        assert "382e332e312e312e3120" == result[-1]

        result = pretty_bytes(bytestream, max_size=None)
        assert "Only dumping 100 bytes." != result[0]
        assert "  43  4f  4d  5f  30  39  30" == result[-1]

        # suffix
        result = pretty_bytes(bytestream, suffix="xxx")
        for line in result:
            assert line[-3:] == "xxx"

    def test_bytesio(self):
        """Test wrap list using bytesio"""
        bytestream = BytesIO()
        bytestream.write(a_associate_rq)
        result = pretty_bytes(bytestream, prefix="", delimiter="", items_per_line=10)
        assert isinstance(result[0], str)


class TestMakeTarget:
    """Tests for utils.make_target()."""

    def test_make_target(self):
        """Context Setup"""
        from contextvars import ContextVar

        foo = ContextVar("foo")
        token = foo.set("foo")

        """Test for ``_config.PASS_CONTEXTVARS = False`` (the default)."""
        assert _config.PASS_CONTEXTVARS is False

        def target_without_context():
            with pytest.raises(LookupError):
                foo.get()

        thread_without_context = Thread(target=make_target(target_without_context))
        thread_without_context.start()
        thread_without_context.join()

        """Test for ``_config.PASS_CONTEXTVARS = True``."""
        _config.PASS_CONTEXTVARS = True

        def target_with_context():
            assert foo.get() == "foo"

        thread_with_context = Thread(target=make_target(target_with_context))
        thread_with_context.start()
        thread_with_context.join()

        _config.PASS_CONTEXTVARS = False

        """Context Teardown"""
        foo.reset(token)


@pytest.fixture
def utf8():
    """Add UTF-8 as a fallback codec"""
    _config.CODECS = ("ascii", "utf8")
    yield
    _config.CODECS = ("ascii",)


class TestSetUID:
    """Tests for utils.set_uid()"""

    def test_str(self):
        """Test str -> UID"""
        uid = set_uid("1.2.3", "foo")
        assert isinstance(uid, UID)
        assert uid == "1.2.3"

    def test_bytes(self):
        """Test bytes -> UID"""
        b = "1.2.3".encode("ascii")
        assert isinstance(b, bytes)
        uid = set_uid(b, "foo")
        assert isinstance(uid, UID)
        assert uid == "1.2.3"

    def test_bytes_decoding_error(self, caplog):
        """Test invalid bytes raises exception"""
        b = BOM_UTF32_LE + "1.2.3".encode("utf_32_le")
        assert isinstance(b, bytes)
        msg = (
            r"Unable to decode 'FF FE 00 00 31 00 00 00 2E 00 00 00 32 00 00 "
            r"00 2E 00 00 00 33 00 00 00' using the ascii codec\(s\)"
        )
        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            with pytest.raises(ValueError, match=msg):
                set_uid(b, "foo")

            assert ("'ascii' codec can't decode byte 0xff in position 0") in caplog.text

    def test_uid(self):
        """Test UID -> UID"""
        uid = set_uid(UID("1.2.3"), "foo")
        assert isinstance(uid, UID)
        assert uid == "1.2.3"

    def test_invalid_raises(self, caplog):
        """Test invalid UID raises exception"""
        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            bad = "abc" * 22
            msg = f"Invalid 'foo' value '{bad}' - must not exceed 64 characters"
            with pytest.raises(ValueError, match=msg):
                uid = set_uid(bad, "foo")
                assert isinstance(uid, UID)
                assert uid == bad

            assert msg in caplog.text

    def test_no_validation(self, caplog):
        """Test skipping validation"""
        with caplog.at_level(logging.WARNING, logger="pynetdicom"):
            uid = set_uid("abc" * 22, "foo", validate=False)
            assert isinstance(uid, UID)
            assert uid == "abc" * 22

            assert not caplog.text

    def test_valid_non_conformant_warns(self, caplog):
        """Test a valid but non-conformant UID warns"""
        with caplog.at_level(logging.WARNING, logger="pynetdicom"):
            uid = set_uid("1.2.03", "foo")
            assert isinstance(uid, UID)
            assert uid == "1.2.03"

            assert "Non-conformant 'foo' value '1.2.03'" in caplog.text

    def test_none_allowed(self):
        """Test None -> None"""
        uid = set_uid(None, "foo", allow_none=True)
        assert uid is None

    def test_none_disallowed(self):
        """Test None raises exception"""
        msg = "'foo' must be str, bytes or UID, not 'NoneType'"
        with pytest.raises(TypeError, match=msg):
            set_uid(None, "foo", allow_none=False)

    def test_empty_passthrough(self):
        """Test an empty value passes through validation"""
        uid = set_uid("", "foo")
        assert uid == UID("")

        uid = set_uid(b"", "foo")
        assert uid == UID("")

        uid = set_uid(UID(""), "foo")
        assert uid == UID("")

    def test_empty_raises(self):
        """Test empty raises exception"""
        msg = r"Invalid 'foo' value - must not be an empty str"
        with pytest.raises(ValueError, match=msg):
            set_uid("", "foo", allow_empty=False)


class TestDecodeBytes:
    """Tests for utils.decode_bytes"""

    def test_decoding_error(self, caplog):
        """Test decoding error raises and logs"""
        b = BOM_UTF32_LE + "1.2.3".encode("utf_32_le")
        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            msg = (
                r"Unable to decode 'FF FE 00 00 31 00 00 00 2E 00 00 00 32 "
                r"00 00 00 2E 00 00 00 33 00 00 00' using the ascii codec\(s\)"
            )
            with pytest.raises(ValueError, match=msg):
                decode_bytes(b)

            assert ("'ascii' codec can't decode byte 0xff in position 0") in caplog.text

    def test_decoding_error_fallback(self, caplog, utf8):
        """Test decoding error raises and logs"""
        b = BOM_UTF32_LE + "1.2.3".encode("utf_32_le")
        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            msg = (
                r"Unable to decode 'FF FE 00 00 31 00 00 00 2E 00 00 00 32 "
                r"00 00 00 2E 00 00 00 33 00 00 00' using the ascii, "
                r"utf8 codec\(s\)"
            )
            with pytest.raises(ValueError, match=msg):
                decode_bytes(b)

            assert ("'ascii' codec can't decode byte 0xff in position 0") in caplog.text
            assert ("'utf-8' codec can't decode byte 0xff in position 0") in caplog.text



#!/usr/bin/env python
"""A QR Get SCU application.

For sending Query/Retrieve (QR) C-GET requests to a QR Get SCP.
"""

import argparse
import sys

from pynetdicom import (
    AE,
    build_role,
    evt,
    StoragePresentationContexts,
)
from pynetdicom.apps.common import setup_logging, create_dataset, handle_store
from pynetdicom._globals import DEFAULT_MAX_LENGTH
from pynetdicom.pdu_primitives import SOPClassExtendedNegotiation
from pynetdicom.sop_class import (
    PatientRootQueryRetrieveInformationModelGet,
    StudyRootQueryRetrieveInformationModelGet,
    PatientStudyOnlyQueryRetrieveInformationModelGet,
    EncapsulatedSTLStorage,
    EncapsulatedOBJStorage,
    EncapsulatedMTLStorage,
)


__version__ = "0.4.0"


def _setup_argparser():
    """Setup the command line arguments"""
    # Description
    parser = argparse.ArgumentParser(
        description=(
            "The getscu application implements a Service Class User "
            "(SCU) for the Query/Retrieve (QR) Service Class. getscu only "
            "supports query functionality using the C-GET message. It sends "
            "query keys to an SCP and waits for a response. The application "
            "can be used to test SCPs of the QR Service Class."
        ),
        usage="getscu [options] addr port",
    )

    # Parameters
    req_opts = parser.add_argument_group("Parameters")
    req_opts.add_argument(
        "addr", help="TCP/IP address or hostname of DICOM peer", type=str
    )
    req_opts.add_argument("port", help="TCP/IP port number of peer", type=int)

    # General Options
    gen_opts = parser.add_argument_group("General Options")
    gen_opts.add_argument(
        "--version", help="print version information and exit", action="store_true"
    )
    output = gen_opts.add_mutually_exclusive_group()
    output.add_argument(
        "-q",
        "--quiet",
        help="quiet mode, print no warnings and errors",
        action="store_const",
        dest="log_type",
        const="q",
    )
    output.add_argument(
        "-v",
        "--verbose",
        help="verbose mode, print processing details",
        action="store_const",
        dest="log_type",
        const="v",
    )
    output.add_argument(
        "-d",
        "--debug",
        help="debug mode, print debug information",
        action="store_const",
        dest="log_type",
        const="d",
    )
    gen_opts.add_argument(
        "-ll",
        "--log-level",
        metavar="[l]",
        help=("use level l for the logger (critical, error, warn, info, debug)"),
        type=str,
        choices=["critical", "error", "warn", "info", "debug"],
    )
    parser.set_defaults(log_type="v")

    # Network Options
    net_opts = parser.add_argument_group("Network Options")
    net_opts.add_argument(
        "-aet",
        "--calling-aet",
        metavar="[a]etitle",
        help="set my calling AE title (default: GETSCU)",
        type=str,
        default="GETSCU",
    )
    net_opts.add_argument(
        "-aec",
        "--called-aet",
        metavar="[a]etitle",
        help="set called AE title of peer (default: ANY-SCP)",
        type=str,
        default="ANY-SCP",
    )
    net_opts.add_argument(
        "-ta",
        "--acse-timeout",
        metavar="[s]econds",
        help="timeout for ACSE messages (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-td",
        "--dimse-timeout",
        metavar="[s]econds",
        help="timeout for DIMSE messages (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-tn",
        "--network-timeout",
        metavar="[s]econds",
        help="timeout for the network (default: 30 s)",
        type=float,
        default=30,
    )
    net_opts.add_argument(
        "-pdu",
        "--max-pdu",
        metavar="[n]umber of bytes",
        help=(
            f"set max receive pdu to n bytes (0 for unlimited, "
            f"default: {DEFAULT_MAX_LENGTH})"
        ),
        type=int,
        default=DEFAULT_MAX_LENGTH,
    )

    # Query information model choices
    qr_group = parser.add_argument_group("Query Information Model Options")
    qr_model = qr_group.add_mutually_exclusive_group()
    qr_model.add_argument(
        "-P",
        "--patient",
        help="use patient root information model (default)",
        action="store_true",
    )
    qr_model.add_argument(
        "-S", "--study", help="use study root information model", action="store_true"
    )
    qr_model.add_argument(
        "-O",
        "--psonly",
        help="use patient/study only information model",
        action="store_true",
    )

    # Query Options
    qr_query = parser.add_argument_group("Query Options")
    qr_query.add_argument(
        "-k",
        "--keyword",
        metavar="[k]eyword: (gggg,eeee)=str, keyword=str",
        help=(
            "add or override a query element using either an element tag as "
            "(group,element) or the element's keyword (such as PatientName)"
        ),
        type=str,
        action="append",
    )
    qr_query.add_argument(
        "-f",
        "--file",
        metavar="path to [f]ile",
        help=(
            "use a DICOM file as the query dataset, if "
            "used with -k then the elements will be added to or overwrite "
            "those present in the file"
        ),
        type=str,
    )

    # Extended Negotiation Options
    ext_neg = parser.add_argument_group("Extended Negotiation Options")
    ext_neg.add_argument(
        "--relational-retrieval",
        help="request the use of relational retrieval",
        action="store_true",
    )
    ext_neg.add_argument(
        "--enhanced-conversion",
        help="request the use of enhanced multi-frame image conversion",
        action="store_true",
    )

    # Output Options
    out_opts = parser.add_argument_group("Output Options")
    out_opts.add_argument(
        "-od",
        "--output-directory",
        metavar="[d]irectory",
        help="write received objects to directory d",
        type=str,
    )
    out_opts.add_argument(
        "--ignore", help="receive data but don't store it", action="store_true"
    )

    ns = parser.parse_args()
    if ns.version:
        pass
    elif not bool(ns.file) and not bool(ns.keyword):
        parser.error("-f and/or -k must be specified")

    return ns


def main(args=None):
    """Run the application."""
    if args is not None:
        sys.argv = args

    args = _setup_argparser()

    if args.version:
        print(f"getscu.py v{__version__}")
        sys.exit()

    APP_LOGGER = setup_logging(args, "getscu")
    APP_LOGGER.debug(f"getscu.py v{__version__}")
    APP_LOGGER.debug("")

    # Create query (identifier) dataset
    try:
        # If you're looking at this to see how QR Get works then `identifier`
        # is a pydicom Dataset instance with your query keys, e.g.:
        #     identifier = Dataset()
        #     identifier.QueryRetrieveLevel = 'PATIENT'
        #     identifier.PatientName = '*'
        identifier = create_dataset(args, APP_LOGGER)
    except Exception as exc:
        APP_LOGGER.exception(exc)
        sys.exit(1)

    # Exclude these SOP Classes
    _exclusion = [
        EncapsulatedSTLStorage,
        EncapsulatedOBJStorage,
        EncapsulatedMTLStorage,
    ]
    store_contexts = [
        cx for cx in StoragePresentationContexts if cx.abstract_syntax not in _exclusion
    ]

    # Create application entity
    # Binding to port 0 lets the OS pick an available port
    ae = AE(ae_title=args.calling_aet)
    ae.acse_timeout = args.acse_timeout
    ae.dimse_timeout = args.dimse_timeout
    ae.network_timeout = args.network_timeout

    # Extended Negotiation - SCP/SCU Role Selection
    ext_neg = []
    ae.add_requested_context(PatientRootQueryRetrieveInformationModelGet)
    ae.add_requested_context(StudyRootQueryRetrieveInformationModelGet)
    ae.add_requested_context(PatientStudyOnlyQueryRetrieveInformationModelGet)
    for cx in store_contexts:
        ae.add_requested_context(cx.abstract_syntax)
        # Add SCP/SCU Role Selection Negotiation to the extended negotiation
        # We want to act as a Storage SCP
        ext_neg.append(build_role(cx.abstract_syntax, scp_role=True))

    if args.study:
        query_model = StudyRootQueryRetrieveInformationModelGet
    elif args.psonly:
        query_model = PatientStudyOnlyQueryRetrieveInformationModelGet
    else:
        query_model = PatientRootQueryRetrieveInformationModelGet

    # Extended Negotiation - SOP Class Extended
    ext_opts = [args.relational_retrieval, args.enhanced_conversion]
    if any(ext_opts):
        app_info = b""
        for option in ext_opts:
            app_info += b"\x01" if option else b"\x00"

        item = SOPClassExtendedNegotiation()
        item.sop_class_uid = query_model
        item.service_class_application_information = app_info
        ext_neg.append(item)

    # Request association with remote
    assoc = ae.associate(
        args.addr,
        args.port,
        ae_title=args.called_aet,
        ext_neg=ext_neg,
        evt_handlers=[(evt.EVT_C_STORE, handle_store, [args, APP_LOGGER])],
        max_pdu=args.max_pdu,
    )

    if assoc.is_established:
        # Send query
        responses = assoc.send_c_get(identifier, query_model)
        for status, rsp_identifier in responses:
            # If `status.Status` is one of the 'Pending' statuses then
            #   `rsp_identifier` is the C-GET response's Identifier dataset
            if status and status.Status in [0xFF00, 0xFF01]:
                # `rsp_identifier` is a pydicom Dataset containing a query
                # response. You may want to do something interesting here...
                pass

        assoc.release()
    else:
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Unit tests for qrscp.py QR move service."""

import logging
import os
import subprocess
import sys
import tempfile
import time

import pytest

try:
    import sqlalchemy

    HAVE_SQLALCHEMY = True
except ImportError:
    HAVE_SQLALCHEMY = False

from pydicom import dcmread, Dataset
from pydicom.uid import (
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    DeflatedExplicitVRLittleEndian,
    ExplicitVRBigEndian,
)

from pynetdicom import AE, evt, debug_logger, DEFAULT_TRANSFER_SYNTAXES, build_role
from pynetdicom.sop_class import (
    CTImageStorage,
    PatientRootQueryRetrieveInformationModelMove,
)


# debug_logger()


APP_DIR = os.path.join(os.path.dirname(__file__), "../")
APP_FILE = os.path.join(APP_DIR, "qrscp", "qrscp.py")
DATA_DIR = os.path.join(APP_DIR, "../", "tests", "dicom_files")


def start_qrscp(args):
    """Start the qrscp.py app and return the process."""
    pargs = [sys.executable, APP_FILE] + [*args]
    return subprocess.Popen(pargs)


def start_qrscp_cli(args):
    """Start the qrscp app using CLI and return the process."""
    pargs = [sys.executable, "-m", "pynetdicom", "qrscp"] + [*args]
    return subprocess.Popen(pargs)


def _send_datasets():
    pargs = [
        sys.executable,
        "-m",
        "pynetdicom",
        "storescu",
        "localhost",
        "11112",
        DATA_DIR,
        "-cx",
    ]
    subprocess.Popen(pargs)


class MoveSCPBase:
    """Tests for qrscp.py"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None
        self.p = None
        self.func = None

        self.tfile = tempfile.NamedTemporaryFile()
        self.db_location = self.tfile.name
        self.instance_location = tempfile.TemporaryDirectory()

        self.startup = 1.0

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

        if self.p:
            self.p.kill()
            self.p.wait(timeout=5)

    def test_basic(self):
        """Test basic operation of the QR get service."""
        self.p = p = self.func(
            [
                "--database-location",
                self.db_location,
                "--instance-location",
                self.instance_location.name,
                "-d",
            ]
        )
        time.sleep(self.startup)
        _send_datasets()
        time.sleep(self.startup)

        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientID = "1CT1"

        datasets = []

        def handle_store(event):
            datasets.append(event.dataset)
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(CTImageStorage)
        scp = ae.start_server(
            ("localhost", 11113),
            block=False,
            evt_handlers=[(evt.EVT_C_STORE, handle_store)],
        )

        model = PatientRootQueryRetrieveInformationModelMove
        ae.add_requested_context(model)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        responses = assoc.send_c_move(query, "STORESCP", model)

        status, ds = next(responses)
        assert status.Status == 0xFF00
        assert ds is None

        status, ds = next(responses)
        assert status.Status == 0x0000
        assert ds is None
        pytest.raises(StopIteration, next, responses)

        assoc.release()

        scp.shutdown()

        p.terminate()
        p.wait()

        assert 1 == len(datasets)
        assert "CompressedSamples^CT1" == datasets[0].PatientName


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestMoveSCP(MoveSCPBase):
    """Tests for qrscp.py"""

    def setup_method(self):
        """Run prior to each test"""
        super().setup_method()
        self.ae = None
        self.p = None
        self.func = start_qrscp


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestMoveSCPCLI(MoveSCPBase):
    """Tests for qrscp using CLI"""

    def setup_method(self):
        """Run prior to each test"""
        super().setup_method()
        self.ae = None
        self.p = None
        self.func = start_qrscp_cli



"""
Implementation of the DIMSE service provider.
"""

from io import BytesIO
import logging
import queue
import threading
from typing import TYPE_CHECKING, Tuple, Optional, Dict, Union, cast

from pynetdicom import evt

# pylint: disable=no-name-in-module
from pynetdicom.dimse_messages import (
    C_STORE_RQ,
    C_STORE_RSP,
    C_FIND_RQ,
    C_FIND_RSP,
    C_GET_RQ,
    C_GET_RSP,
    C_MOVE_RQ,
    C_MOVE_RSP,
    C_ECHO_RQ,
    C_ECHO_RSP,
    C_CANCEL_RQ,
    N_EVENT_REPORT_RQ,
    N_GET_RQ,
    N_SET_RQ,
    N_ACTION_RQ,
    N_CREATE_RQ,
    N_DELETE_RQ,
    N_EVENT_REPORT_RSP,
    N_GET_RSP,
    N_SET_RSP,
    N_ACTION_RSP,
    N_CREATE_RSP,
    N_DELETE_RSP,
    DIMSEMessage,
)

# pylint: enable=no-name-in-module
from pynetdicom.dimse_primitives import (
    C_STORE,
    C_FIND,
    C_GET,
    C_MOVE,
    C_ECHO,
    C_CANCEL,
    N_EVENT_REPORT,
    N_GET,
    N_SET,
    N_ACTION,
    N_CREATE,
    N_DELETE,
    DimsePrimitiveType,
    DimseServiceType,
)
from pynetdicom.utils import make_target

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.association import Association
    from pynetdicom.dul import DULServiceProvider
    from pynetdicom.pdu_primitives import P_DATA


LOGGER = logging.getLogger(__name__)

_RQ_TO_MESSAGE = {
    C_ECHO: C_ECHO_RQ,
    C_STORE: C_STORE_RQ,
    C_FIND: C_FIND_RQ,
    C_MOVE: C_MOVE_RQ,
    C_GET: C_GET_RQ,
    N_EVENT_REPORT: N_EVENT_REPORT_RQ,
    N_GET: N_GET_RQ,
    N_SET: N_SET_RQ,
    N_ACTION: N_ACTION_RQ,
    N_CREATE: N_CREATE_RQ,
    N_DELETE: N_DELETE_RQ,
}
_RSP_TO_MESSAGE = {
    C_ECHO: C_ECHO_RSP,
    C_STORE: C_STORE_RSP,
    C_FIND: C_FIND_RSP,
    C_MOVE: C_MOVE_RSP,
    C_GET: C_GET_RSP,
    C_CANCEL: C_CANCEL_RQ,
    N_EVENT_REPORT: N_EVENT_REPORT_RSP,
    N_GET: N_GET_RSP,
    N_SET: N_SET_RSP,
    N_ACTION: N_ACTION_RSP,
    N_CREATE: N_CREATE_RSP,
    N_DELETE: N_DELETE_RSP,
}

_QueueItem = Union[Tuple[None, None], Tuple[int, DimseServiceType]]


class DIMSEServiceProvider:
    """The DIMSE service provider.

    .. versionchanged:: 1.2

        Added `msg_queue` attribute

    **Messages**

    +----------------+-----------------------+------------------------+
    | Primitive      | Type                  | Message Class          |
    +================+=======================+========================+
    | C-CANCEL       | Request/indication    | C_CANCEL_RQ            |
    +----------------+-----------------------+------------------------+
    | C-ECHO         | Request/indication    | C_ECHO_RQ              |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_ECHO_RSP             |
    +----------------+-----------------------+------------------------+
    | C-FIND         | Request/indication    | C_FIND_RQ              |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_FIND_RSP             |
    +----------------+-----------------------+------------------------+
    | C-GET          | Request/indication    | C_GET_RQ               |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_GET_RSP              |
    +----------------+-----------------------+------------------------+
    | C-MOVE         | Request/indication    | C_MOVE_RQ              |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_MOVE_RSP             |
    +----------------+-----------------------+------------------------+
    | C-STORE        | Request/indication    | C_STORE_RQ             |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | C_STORE_RSP            |
    +----------------+-----------------------+------------------------+
    | N-ACTION       | Request/indication    | N_ACTION_RQ            |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_ACTION_RSP           |
    +----------------+-----------------------+------------------------+
    | N-CREATE       | Request/indication    | N_CREATE_RQ            |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_CREATE_RSP           |
    +----------------+-----------------------+------------------------+
    | N-DELETE       | Request/indication    | N_DELETE_RQ            |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_DELETE_RSP           |
    +----------------+-----------------------+------------------------+
    | N-EVENT-REPORT | Request/indication    | N_EVENT_REPORT_RQ      |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_EVENT_REPORT_RSP     |
    +----------------+-----------------------+------------------------+
    | N-GET          | Request/indication    | N_GET_RQ               |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_GET_RSP              |
    +----------------+-----------------------+------------------------+
    | N-SET          | Request/indication    | N_SET_RQ               |
    |                +-----------------------+------------------------+
    |                | Response/confirmation | N_SET_RSP              |
    +----------------+-----------------------+------------------------+

    Attributes
    ----------
    cancel_rq : dict
        A dict of ``{MessageIDBeingRespondedTo : C_CANCEL}`` messages received.
        The dict is cleared out at the start and end of Service Class
        operations and is limited to a maximum of 10 messages.
    message : dimse_messages.DIMSEMessage or None
        The DIMSE message currently being received.
    msg_queue: queue.queue of dimse_messages.DIMSEMessage
        A queue holding decoded DIMSE Message primitives received from the
        peer, except for C-CANCEL requests.

    References
    ----------

    * DICOM Standard, :dcm:`Part 7<part07/PS3.7.html>`
    """

    def __init__(self, assoc: "Association") -> None:
        """Initialise the DIMSE service provider.

        .. versionchanged:: 1.3

            Class initialisation now only takes `assoc` parameter.

        Parameters
        ----------
        assoc : association.Association
            The association to provide DIMSE services for.
        """
        self._assoc = assoc

        self.cancel_req: Dict[int, C_CANCEL] = {}
        self.message: Optional[DIMSEMessage] = None
        self.msg_queue: "queue.Queue[_QueueItem]" = queue.Queue()

    @property
    def assoc(self) -> "Association":
        """Return the parent :class:`~pynetdicom.association.Association`.

        .. versionadded:: 1.3
        """
        return self._assoc

    @property
    def dimse_timeout(self) -> Optional[float]:
        """Return the DIMSE timeout as numeric or ``None``."""
        return self.assoc.dimse_timeout

    @property
    def dul(self) -> "DULServiceProvider":
        """Return the :class:`~pynetdicom.dul.DULServiceProvider`."""
        return self.assoc.dul

    def get_msg(self, block: bool = False) -> _QueueItem:
        """Get the next available DIMSE message.

        .. versionadded:: 1.2

        .. versionchanged:: 1.5

            Changed to also return ``(None, None)`` if the peer aborts the
            association or the connection is closed.

        Parameters
        ----------
        block : bool
            If ``True`` then the function will block until either a message is
            available or :attr:`~DIMSEServiceProvider.dimse_timeout` expires,
            otherwise non-blocking.

        Returns
        -------
        (int, dimse_messages.DIMSEMessage) or (None, None)
            The next available (*Context ID*, *DIMSE Message*), which is taken
            off the queue, or ``(None, None)`` if the peer has aborted the
            association, the connection is closed or if no messages are
            available within the :attr:`~DIMSEServiceProvider.dimse_timeout`
            period.
        """
        try:
            return self.msg_queue.get(block=block, timeout=self.dimse_timeout)
        except queue.Empty:
            return None, None

    @property
    def maximum_pdu_size(self) -> int:
        """Return the peer's maximum PDU length as :class:`int`."""
        if self.assoc.is_requestor:
            return cast(int, self.assoc.acceptor.maximum_length)

        return cast(int, self.assoc.requestor.maximum_length)

    def peek_msg(self) -> _QueueItem:
        """Return the first message in the message queue or ``None``.

        .. versionadded:: 1.2

        Returns
        -------
        (int, dimse_messages.DIMSEMessage) or (None, None)
            The first (*Context ID*, *Message*) in the queue if one is
            available, otherwise ``(None, None)``. No messages are taken out
            of the queue.
        """
        try:
            return cast(_QueueItem, self.msg_queue.queue[0])
        except (queue.Empty, IndexError):
            return None, None

    def receive_primitive(self, primitive: "P_DATA") -> None:
        """Process a P-DATA primitive received from the remote.

        .. versionadded:: 1.2

        A DIMSE message is split into one or more P-DATA primitives, which
        must be sent in sequential order. While waiting for all the P-DATA
        primitives associated with a message the encoded data is stored in
        :attr:`~DIMSEServiceProvider.message`, which is decoded only when
        complete and converted into a DIMSE Message primitive which is added
        to the :attr:`~DIMSEServiceProvider.msg_queue`.

        This makes it possible to process incoming P-DATA primitives into
        DIMSE messages while a service class implementation is running.

        Parameters
        ----------
        primitive : pdu_primitives.P_DATA
            A P-DATA primitive received from the peer to be processed.
        """
        if self.message is None:
            self.message = DIMSEMessage()

        if self.message.decode_msg(primitive, self.assoc):
            # Trigger event
            evt.trigger(self.assoc, evt.EVT_DIMSE_RECV, {"message": self.message})

            context_id = cast(int, self.message.context_id)
            try:
                d_primitive = self.message.message_to_primitive()
            except Exception as exc:
                LOGGER.error("Received an invalid DIMSE message")
                LOGGER.exception(exc)
                self.dul.event_queue.put("Evt19")
                # Shouldn't need to reset `self.message` as Evt19 will end
                #   the association
                return

            # Keep C-CANCEL requests separate from other messages
            # Only allow up to 10 C-CANCEL requests
            if isinstance(d_primitive, C_CANCEL) and len(self.cancel_req) < 10:
                msg_id = cast(int, d_primitive.MessageIDBeingRespondedTo)
                self.cancel_req[msg_id] = d_primitive
            elif (
                isinstance(d_primitive, N_EVENT_REPORT) and d_primitive.is_valid_request
            ):
                # N-EVENT-REPORT service requests are handled immediately
                # Ugly hack, but would block the DUL otherwise
                t = threading.Thread(
                    target=make_target(self.assoc._serve_request),
                    args=(d_primitive, context_id),
                )
                t.start()
            else:
                self.msg_queue.put((context_id, cast(DimseServiceType, d_primitive)))

            # Fix for memory leak, Issue #41
            #   Reset the DIMSE message, ready for the next one
            self.message.encoded_command_set = BytesIO()
            self.message.data_set = BytesIO()
            self.message._data_set_file = None
            self.message._data_set_path = None
            self.message = None

    def send_msg(self, primitive: DimsePrimitiveType, context_id: int) -> None:
        """Encode and send a DIMSE-C or DIMSE-N message to the peer AE.

        Parameters
        ----------
        primitive : dimse_primitives DIMSE Primitive class
            The DIMSE message primitive to send to the peer.
        context_id : int
            The ID of the presentation context that the message is to be
            sent under.
        """
        if primitive.MessageIDBeingRespondedTo is None:
            dimse_msg = _RQ_TO_MESSAGE[primitive.__class__]()  # type: ignore
        else:
            dimse_msg = _RSP_TO_MESSAGE[primitive.__class__]()

        # Convert DIMSE primitive to DIMSE Message
        dimse_msg.primitive_to_message(primitive)
        dimse_msg.context_id = context_id

        # Trigger event
        evt.trigger(self.assoc, evt.EVT_DIMSE_SENT, {"message": dimse_msg})

        # Split the full messages into P-DATA chunks,
        #   each below the max_pdu size
        for pdata in dimse_msg.encode_msg(context_id, self.maximum_pdu_size):
            self.dul.send_pdu(pdata)
